"""Toolkit Module.

This module contains a number of classes designed to support development of WPF GUI-based toolkits for AEDT
applications. The primary classes are:

    WPFToolkit             : Base Class to support UI Toolkit development
    WPFToolkitSettings     : Class to support UI settings for testing purposes
    ToolkitBuilder          : Class to support generation of a zip package with all needed dependencies

Examples
--------

**Example 1**
A toolkit is built by deriving a class called ApplicationWindow from WPFToolkit.
Parallel to this file, an xaml file should be present in the same directory.


>>> from pyaedt import Maxwell2d
>>> from pyaedt.generic.toolkit import WPFToolkit, launch

>>> class ApplicationWindow(WPFToolkit):
>>>
>>>     def __init__(self):
>>>         WPFToolkit.__init__(self, toolkit_file=__file__, aedt_design=Maxwell2d(), parent_design_name=None)
>>>         # Copy Xaml template
>>>         self.copy_xaml_template()

>>>         #Edit the UI
>>>        self.add_label("label1", "Input Parameter 1", 10, 10)
>>>        self.add_text_box(name="_input1", x_pos=150, y_pos=10, callback_method=self.validate_positive_float,
>>>                          callback_action='LostFocus')
>>>        self.add_label("label2", "Design Name", 10, 50)
>>>        self.add_text_box(name="_input2", x_pos=150, y_pos=50, callback_method=self.validate_string_no_spaces,
>>>                          callback_action='LostFocus')
>>>        self.add_check_box(name="_check1", content="Save Project", x_pos=150, y_pos=80,
>>>                           callback_method=self.enable_checkbox, callback_action="Checked")
>>>        self.add_combo_box(name="_combo1", x_pos=150, y_pos=120, callback_method=self.print_design_name)
>>>        self.add_button("run_method", "Click To Run", x_pos=300, y_pos=300, callback_method=self.custom_callback)
>>>        self.add_label("label3", "Design Type", 10, 150)
>>>        self.add_text_box(name="_input3", x_pos=150, y_pos=150)
>>>
>>>         #Launch the UI
>>>        self.launch_gui()
>>>
>>>         # Setup Additional Callbacks
>>>         self.add_combo_items("_combo1", self.aedtdesign.design_list)

>>>
>>>         # Get Values from UI
>>>         value1 = self.ui.float_value("_input1")
>>>         value2 = self.ui.text_value("_input2")
>>>
>>>     def custom_callback(self, sender, e):
>>>         self.aedtdesign.design_name = self.ui.text_value("_input2")
>>>         self.aedtdesign["param"] = self.ui.float_value("_input1")
>>>         print(self.ui.float_value("_input1"))
>>>         print(self.ui.text_value("_input2"))
>>>
>>>     def print_design_name(self, sender, e):
>>>         self.set_text_value("_input3", self.aedtdesign.solution_type)
>>>
>>>    def enable_checkbox(self, sender, e):
>>>        print("Enabled")
>>>
>>>    def disable_checkbox(self, sender, e):
>>>        print("Disabled")
>>> # Launch the toolkit
>>> if __name__ == '__main__':
>>>     launch(__file__, specified_version="2021.2", new_desktop_session=False, autosave=False)
"""
import json
import os
import shutil
import sys
from datetime import datetime
from zipfile import ZIP_DEFLATED
from zipfile import ZipFile

import clr

from pyaedt import is_ironpython
from pyaedt.desktop import Desktop
from pyaedt.generic.general_methods import pyaedt_function_handler

if os.name == "posix" and is_ironpython:
    import subprocessdotnet as subprocess
else:
    import subprocess

if is_ironpython:
    clr.AddReference("PresentationFramework")
    clr.AddReference("PresentationCore")
    clr.AddReference("System.Windows")
    clr.AddReference("System.Windows.Forms")
    clr.AddReference("System.Drawing")
else:
    clr.AddReference("System.Xml")
    clr.AddReference("PresentationFramework, Version=3.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35")
    clr.AddReference("PresentationCore, Version=3.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35")
    clr.AddReference("System.Windows.Forms")
    clr.AddReference("System")
    from System.IO import StreamReader
    from System.Windows.Markup import XamlReader

    clr.AddReference("System.Windows")

from System.Threading import Thread, ThreadStart, ApartmentState
from System.Windows import LogicalTreeHelper
from System.Windows import Window, Application, Visibility, Input, Thickness
from System.Windows.Forms import Form, ListBox, Button, MessageBox, MessageBoxIcon, MessageBoxButtons
from System.Windows.Forms import (
    FormBorderStyle,
    StatusBar,
    SelectionMode,
    DialogResult,
    FolderBrowserDialog,
    OpenFileDialog,
)
from System.Windows.Media import Brushes
from System.Drawing import Size, Point

from System.Windows.Media.Imaging import BitmapImage, BitmapCacheOption, BitmapCreateOptions
from System import Uri, UriKind, Environment
from System.Windows.Media.Imaging import BitmapImage


@pyaedt_function_handler()
def select_file(initial_dir=None, filter=None):
    """Opens File Dialog and select a file.

    Parameters
    ----------
    initial_dir : str
        Initial dir where to look for file.
    filter : str
        Filter for file search.

    Returns
    -------
    str
        File full path.
    """
    browser = OpenFileDialog()
    browser.Title = "Select a file with optional filter and directory"
    if initial_dir:
        browser.InitialDirectory = initial_dir
    if filter:
        browser.Filter = filter
    ret_val = browser.ShowDialog()
    if ret_val == DialogResult.OK:
        return browser.FileName
    else:
        return None


@pyaedt_function_handler()
def select_directory(initial_dir=None, description=None):
    """Opens File Dialog and select a directory.

    Parameters
    ----------
    initial_dir : str
        Initial dir where to look for.
    description : str
        Dialog description.

    Returns
    -------
    str
        Directory full path.
    """
    browser = FolderBrowserDialog()

    if initial_dir:
        browser.RootFolder = Environment.SpecialFolder.Desktop
        browser.SelectedPath = initial_dir
    if description:
        browser.Description = description
    else:
        browser.Description = "Select an empty folder to create the toolkit"

    test = browser.ShowDialog()
    if test == DialogResult.OK:
        return browser.SelectedPath
    else:
        return None


@pyaedt_function_handler()
def copy_files_mkdir(root, files_in_subdir):
    """Copies all files from source to destination in a root path.

    Parameters
    ----------
    root : str
        Root Path where copy new files.
    files_in_subdir : dict
        Dictionary of source and destinations files. key is source, value is destination

    Returns
    -------
    bool
    """
    if not os.path.exists(root):
        os.mkdir(root)
    for src_file, dest_file in files_in_subdir.items():
        shutil.copyfile(src_file, dest_file)
    return True


@pyaedt_function_handler()
def launch(
    workflow_module, specified_version=None, new_desktop_session=True, autosave=False, close_desktop_on_exit=False
):
    """Launches the ApplicationWindow class for the given module name which must lie within the path scope of
        the calling toolkit. Optionally the version can be specified in the form 20xx.y, e.g. 2021.2,
        IronPython calls the run_application function of the ApplicationThread class.
        For CPython, a Thread with Apartment State is generated to adhere to requirements of System.Windows


    Parameters
    ----------
    workflow_module : __file__
    name of the module file containing the application.
    specified_version : str
        Specified version of Desktop to launch
    new_desktop_session : bool
        Define if new Desktop session has always to be opened.
    autosave : bool
        Enable or disable autosave.
    close_desktop_on_exit : bool
        Define if Desktop has to be closed on exit.
    """
    sys.path.append(os.path.dirname(workflow_module))
    app = ApplicationThread(workflow_module, specified_version, new_desktop_session, autosave, close_desktop_on_exit)

    if sys.implementation.name != "ironpython":
        thread = Thread(ThreadStart(app.run_application))
        thread.SetApartmentState(ApartmentState.STA)
        thread.Start()
        thread.Join()
    else:
        if specified_version:
            app.run_application()
        else:
            app.open_form()


launch_script = """
import os
import sys

try:
    specified_version = sys.argv[1]
except IndexError:
    specified_version = None

# Ensure that this launch directory is included to the search path (needed for IronPython under AEDT)
toolkit_directory = os.path.abspath(os.path.dirname(__file__))
toolkit_lib_directory = os.path.join(toolkit_directory, 'lib')
sys.path.append(toolkit_directory)
sys.path.append(toolkit_lib_directory)

from lib.AEDTLib.Toolkit import launch
launch('{}', version=specified_version)
"""


@pyaedt_function_handler()
def message_box(text, caption=None, buttons=None, icon=None):
    """Displays Message Box.

    Parameters
    ----------
    text : str
        Main text of the message
    caption : str
        Caption of the window
    buttons : str
        Button type string (OK, OKCancel, YesNoCancel, YesNo).
    icon : str
        Valid icon type string.  Asterisk, Error, Exclamation, Hand, Information, None, Question, Stop, Warning
    """

    if not icon:
        icon_object = getattr(MessageBoxIcon, "Information")
    else:
        icon_object = getattr(MessageBoxIcon, "Information")

    if not caption:
        caption = ""

    if not buttons:
        buttons_object = MessageBoxButtons.OK
    else:
        buttons_object = getattr(MessageBoxButtons, buttons)

    result = MessageBox.Show(text, caption, buttons_object, icon_object)
    return result


class ToolkitBuilder:
    """Helps to create a deployable zip file of the application and any dependencies that are not available via pip.

    Parameters
    ----------
        local_path : str
            Path of the top-level toolkit files containing the *.py and *.xaml files.
            Optionally a file located in the toolkit directory (typically __file__)
            can also be given.

        app_name : str, optional
            Name of the toolkit (optional).
            If not defined, then the name of the parent directory is used.

    Examples
    --------

    Build file placed in the Toolkit directory

    >>> from pyaedt.generic.toolkit import ToolkitBuilder
    >>>  bld = ToolkitBuilder(__file__)
    >>>  bld.copy_from_local(extension=['py', 'xaml'])
    >>>  bld.copy_from_repo(sub_dir='pyaedt')
    >>>  bld.zip_archive()
    """

    def __init__(self, local_path, app_name=None):
        """Instantiates a ToolkitBuilder object. This object manages the packaging of the WPF GUI to ensure
        that any dependencies not available via pip are stored with the toolkit deployable asset.
        """

        # if the local_path is defined as a file, take the path of that file
        if os.path.isfile(local_path):
            self.local_path = os.path.dirname(os.path.realpath(local_path))
        else:
            self.local_path = local_path

        if app_name:
            self.py_name = app_name
        else:
            self.py_name = os.path.basename(os.path.dirname(os.path.realpath(local_path)))

        self.toolkit_path = os.path.join(self.local_path, "..")

        # Extract the commit ID from the global repository
        self.global_lib_path = os.path.abspath(os.path.join(self.local_path, "..", "..", ".."))
        os.chdir(self.global_lib_path)
        command = ["git", "rev-parse", "HEAD"]
        sh = False
        res = subprocess.check_output(command, shell=sh).rstrip()

        # Create a directory called 'build' in the xml directory
        self.build_path = os.path.join(self.local_path, ".build")
        if not os.path.isdir(self.build_path):
            os.mkdir(self.build_path)

        # Create a drectory named by the datetime (remove contents if already exists)
        self.build_name = "{0}-{date:%Y%m%d_%H%M}".format(self.py_name, date=datetime.now())
        self.commit_path = os.path.join(self.build_path, self.build_name)
        self.commit_lib_path = os.path.join(self.commit_path, "lib")
        if os.path.isdir(self.commit_path):
            shutil.rmtree(self.commit_path)
        os.mkdir(self.commit_path)
        os.mkdir(self.commit_lib_path)

        # Add the commit hash to the module __init__.py files
        init_files = []
        init_files.append(os.path.join(self.commit_lib_path, "__init__.py"))
        for file in init_files:
            with open(file, "w") as f:
                f.write("#{0}\n".format(res))

        # Create a launch file
        script_file = os.path.join(self.commit_path, "launch.py")
        with open(script_file, "w") as f:
            f.write(launch_script.format(self.py_name))

    def copy_from_local(self, extension=None, ignore_dir=None):
        """Copies recursively all files in the local directory and all subdirectories of a given list of
            extensions

        :param extension: list of extensions to be copied, e.g. ['py', 'xaml']
        :param ignore_dir: boolean: ignore directories starting with "." or "_"
        """
        self.copy_from_repo(self.local_path, extension=extension, ignore_dir=ignore_dir)

    def copy_from_repo(self, root_dir=None, sub_dir=None, extension=None, ignore_dir=None):
        """Copies recursively all files from a specified root directory and all subdirectories of a given list of
            extensions

        :param root_dir: optional root directory to copy from, default is the AnsysAutomation repository
        :param sub_dir: optional specific sub-directory within the root directory
        :param extension: list of extensions to be copied, e.g. ['py', 'xaml']
        :param ignore_dir: boolean: ignore directories starting with "." or "_"
        """
        if not extension:
            extension = ["py"]
        elif isinstance(extension, str):
            extension = [extension]
        assert isinstance(extension, list), "Extension input parameter must be a string or a list"

        if not ignore_dir:
            ignore_dir = [".", "_"]
        elif isinstance(ignore_dir, str):
            extension = [ignore_dir]
        assert isinstance(ignore_dir, list), "Extension input parameter must be a string or a list"

        if not root_dir:
            root_dir = self.global_lib_path
        walk_dir = root_dir
        if sub_dir:
            if isinstance(sub_dir, str):
                sub_dir = [sub_dir]
            assert isinstance(sub_dir, list), "sub_dir must be a list or a string"
            for x in sub_dir:
                walk_dir = os.path.join(walk_dir, x)

        ignored_paths = []
        for root, dir, files in os.walk(walk_dir):
            ignore_path = False
            if ignore_dir:
                for x in ignore_dir:
                    if os.path.basename(root).startswith(x):
                        ignored_paths.append(root)
                        ignore_path = True
                        break
                for x in ignored_paths:
                    if root.startswith(x):
                        ignore_path = True
                        break
            if not ignore_path:
                files_in_subdir = {}
                dest_dir = root.replace(root_dir, self.commit_lib_path)
                for name in files:
                    for ext in extension:
                        if name.endswith((ext)):
                            src_file = os.path.join(root, name)
                            dst_file = src_file.replace(root_dir, self.commit_lib_path)
                            files_in_subdir[src_file] = dst_file
                            break
                copy_files_mkdir(dest_dir, files_in_subdir)

    def zip_archive(self):
        """Zip the collected data in the build path"""

        zip_archive = os.path.join(self.build_path, self.build_name + ".zip")
        with ZipFile(zip_archive, "w", ZIP_DEFLATED) as myzip:
            for root, dirs, files in os.walk(self.commit_path):
                for filename in files:
                    abs_path = os.path.join(root, filename)
                    rel_path = os.path.relpath(abs_path, self.commit_path)
                    myzip.write(abs_path, arcname=rel_path)

        # remove the commit directory
        shutil.rmtree(self.commit_path)


class ApplicationThread:
    """Class to allow data to be passed to the run_application member function in a new thread.
    The run_application function connects to Desktop and instantiates an ApplicationWindow
    object of the calling module
    """

    def __init__(self, workflow_module, version, new_desktop_session=True, autosave=False, close_desktop_on_exit=False):
        self.workflow_module = os.path.basename(workflow_module).replace(".py", "")
        self.version = version
        self.autosave = autosave
        self.new_desktop = new_desktop_session
        self.close_desktop_on_exit = close_desktop_on_exit

    @pyaedt_function_handler()
    def run_application(self):
        """Starts the application and run WPF."""
        d = Desktop(self.version, new_desktop_session=self.new_desktop)
        if self.autosave:
            d.enable_autosave()
        else:
            d.disable_autosave()
        form_object = __import__(self.workflow_module)
        with form_object.ApplicationWindow() as form:
            form.display()
            print(self.close_desktop_on_exit)
            if self.close_desktop_on_exit:
                d.release_desktop(True, True)
            else:
                d.release_desktop(False, False)

    @pyaedt_function_handler()
    def open_form(self):
        """Opens the Application Windows Form."""
        form_object = __import__(self.workflow_module)
        with form_object.ApplicationWindow() as form:
            form.display()


# Manages the settings data for the toolkit
class WPFToolkitSettings:
    """This class provides a minimal implementation of the Toolkit providing assess to the
    settings file and allowing to call the toolkit_functionality without the
    GUI to speed up debugging (or potentially to deploy the function in batch mode.

    Examples
    --------

    Typical usage looks like this

    >>> with Desktop(version="2021.2") as d:
    >>>    app = WPFToolkitSettings(toolkit_file=__file__, aedt_app=Maxwell2d())
    >>>    app.settings_data = {'param 1': 3, 'param 2': 'house' }
    >>>    toolkit_function(app)
    """

    def __init__(self, aedtdesign=None, working_directory=None, toolkit_name=None):
        if aedtdesign:
            self._parent = aedtdesign
            self._working = None
        elif working_directory:
            assert os.path.exists(working_directory), "Working Directory {} does not exist"
            self._working = working_directory
        else:
            self._parent = None
            self._working = None
            # raise Exception("Toolkit name/file not defined")

        self.toolkit_name = toolkit_name

    @property
    def settings_file(self):
        """Settings json file path."""
        return os.path.join(self.settings_path, self.toolkit_name + "_Settings.json")

    @property
    def results_path(self):
        """Results folder path."""
        return os.path.join(self._parent.working_directory, "_results")

    @property
    def data_path(self):
        """Data folder path."""
        return os.path.join(self._working, "_data")

    @property
    def local_path(self):
        """Working folder path."""
        if self._working:
            return self._working
        else:
            return self._parent.working_directory

    @property
    def settings_path(self):
        """Working directory of the parent design - checks the current design settings
        file for the key "parent" to find the name of the parent design
        """
        my_path = self.local_path
        my_settings_file = os.path.join(self.local_path, self.toolkit_name + "_Settings.json")
        if os.path.exists(my_settings_file):
            settings_data = self.read_settings_file(my_settings_file)
            if "parent" in settings_data:
                if settings_data["parent"]:
                    my_dir = os.path.basename(my_path)
                    my_path = my_path.replace(my_dir, settings_data["parent"])
        return my_path

    @pyaedt_function_handler()
    def read_settings_file(self, filename):
        """Read the json file and returns dictionary."""
        with open(filename, "r") as f:
            try:
                settings_data = json.load(f)
            except ValueError:
                try:
                    msg_string = "Invalid json file {0} will be overwritten.".format(filename)
                    self._parent.logger.warning(msg_string)
                except:
                    pass
                return None
        return settings_data

    @property
    def settings_data(self):
        """Reads the json file and returns dictionary."""
        settings_file = self.settings_file
        if os.path.exists(settings_file):
            return self.read_settings_file(self.settings_file)
        else:
            return {"parent": None}

    @pyaedt_function_handler()
    def append_toolkit_dir(self):
        """Appends toolkit directory to sys.path."""
        assert os.path.exists(self.settings_file), "Settings File not defined!"
        try:
            toolkit_dir = self.settings_data["_toolkit_dir"]
            sys.path.append(toolkit_dir)
        except KeyError:
            pass


class ListBoxForm(Form):
    @property
    def ResultOK(self):
        if self.DialogResult == DialogResult.OK:
            return True
        else:
            return False

    def __init__(self, list_items, default_item=None):

        self.MinimizeBox = False
        self.MaximizeBox = False
        self.Text = "Select Export Objects"

        self.lb = ListBox()
        self.lb.Parent = self
        for item in list_items:
            self.lb.Items.Add(item)
        self.lb.Location = Point(10, 10)
        self.lb.Size = Size(200, 250)
        self.lb.SelectedIndexChanged += self.OnChanged
        self.lb.SelectionMode = SelectionMode.MultiExtended
        self.sb = StatusBar()
        self.sb.Parent = self
        self.Size = Size(300, 350)
        self.CenterToScreen()

        # Define the border style of the form to a dialog box.
        self.FormBorderStyle = FormBorderStyle.FixedDialog
        button1 = Button()
        button2 = Button()

        # Set the text of button1 to "OK".
        button1.Text = "OK"
        button2.Text = "Cancel"

        button1.Location = Point(10, self.lb.Bottom + 5)
        button2.Location = Point(button1.Left + button1.Width + 10, button1.Top)

        button1.DialogResult = DialogResult.OK
        button2.DialogResult = DialogResult.Cancel

        self.AcceptButton = button1
        self.CancelButton = button2

        # Add buttons to the form.
        self.Controls.Add(button1)
        self.Controls.Add(button2)

        if default_item:
            if default_item in self.lb.Items:
                index_of_default_item = self.lb.Items.IndexOf(default_item)
                self.lb.SelectedIndex = index_of_default_item

    def OnChanged(self, sender, event):
        self.sb.Text = ";".join(sender.SelectedItems)


class UIObjectGetter:
    """Child class of a parent supplying a get_ui_object function returning a WPF GUI object.
    The class implementa a __getitem__ method to allow for accessing the GUI objects from the parent class
    through the syntax:
    parent.UIObjectGetter[ui_object_name]

    Additionally methods are provided to extract specific data types from the GUI object:
    float_value
    text_value
    """

    def __init__(self, parent):
        self.parent = parent

    def __getitem__(self, ui_object_name):
        return self.parent.get_ui_object(ui_object_name)

    @pyaedt_function_handler()
    def float_value(self, ui_object_name):
        """Converts the text entry to a float and return the value. If the string is empty, return 0

        Parameters
        ----------
        ui_object_name : str
            Name of UI object.

        Returns
        -------
        float
        """
        text_value = self[ui_object_name].Text
        if not text_value:
            return 0
        else:
            return float(text_value)

    def text_value(self, ui_object_name):
        """Converts the text entry to a string and return the value. If the string is empty, return 0

        Parameters
        ----------
        ui_object_name : str
            Name of UI object.

        Returns
        -------
        str
        """
        return self[ui_object_name].Text


class WPFToolkit(Window):
    """This class provides a base class allowing the creation of a WPF-GUI-Based toolkit for AEDT.
    This class provides functionality for launching the GUI, reading and writing to settings
    files, path handling and error handling.
    Data validation functions are also provided as callbacks from the WPF GUI
    """

    def __init__(self, toolkit_file, aedt_design=None, parent_design_name=None):

        self.toolkit_file = toolkit_file
        self._aedtdesign = None
        self.aedtdesign = aedt_design
        self.parent_design_name = parent_design_name
        self.window = None
        self.ui = UIObjectGetter(self)
        my_path = os.path.abspath(os.path.dirname(__file__))
        self.toolkit_directory = os.path.abspath(os.path.dirname(toolkit_file))
        self.aedtlib_directory = os.path.abspath(os.path.join(my_path, ".."))
        sys.path.append(self.toolkit_directory)
        sys.path.append(self.aedtlib_directory)
        self.image_path = os.path.join(self.aedtlib_directory, "misc")

        self.dsoconfigfile = os.path.join(self.toolkit_directory, "dso.cfg")

        # Read existing settings and update the library path
        self._read_and_synch_settings_file()
        self._callbacks = []
        # LOCAL_INSTALL = self.aedtdesign.odesktop.GetExeDir()
        # self.desktopjob = os.path.join(LOCAL_INSTALL, "desktopjob.exe")

    @property
    def aedtdesign(self):
        """Return Aedt Object."""
        return self._aedtdesign

    @aedtdesign.setter
    def aedtdesign(self, design):
        my_path = os.path.abspath(os.path.dirname(__file__))
        self._aedtdesign = design
        self.toolkit_name = os.path.basename(self.toolkit_file).replace(".py", "")
        if self._aedtdesign:
            self.settings_manager = WPFToolkitSettings(aedtdesign=self._aedtdesign, toolkit_name=self.toolkit_name)
            self._parent_design_name = self._aedtdesign.design_name
        else:
            self.settings_manager = WPFToolkitSettings(working_directory=my_path, toolkit_name=self.toolkit_name)
            self._parent_design_name = None

    @property
    def parent_design_name(self):
        """Aedt Design Name."""

        if self.aedtdesign:
            self._parent_design_name = self.aedtdesign.design_name
        else:
            self._parent_design_name = None
        return self._parent_design_name

    @parent_design_name.setter
    def parent_design_name(self, design_name=None):
        if not self.aedtdesign:
            self._parent_design_name = None
        elif design_name:
            self._parent_design_name = design_name
            if not design_name in self.aedtdesign.design_list:
                orig_design_name = self.aedtdesign.design_name
                if self._parent_design_name != orig_design_name:
                    self._write_parent_link()
                    self.aedtdesign.duplicate_design(self._parent_design_name)
                    self.aedtdesign.save_project()
            else:
                self.aedtdesign.set_active_design(design_name)
        else:
            self._parent_design_name = self.aedtdesign.design_name

    @property
    def results_path(self):
        """Results folder path."""
        return os.path.join(self.aedtdesign.working_directory, "_results")

    @property
    def data_path(self):
        """Data folder path."""
        return os.path.join(self.aedtdesign.working_directory, "_data")

    @property
    def local_path(self):
        """Local folder path."""
        return self.settings_manager.local_path

    @property
    def settings_file(self):
        """Settings json file path."""
        try:
            return self.settings_manager.settings_file
        except:
            return ""

    @property
    def settings_data(self):
        """Settings json file data."""
        try:
            return self.settings_manager.settings_data
        except:
            return None

    @property
    def local_settings_file(self):
        """Settings json file path."""
        return self.settings_manager.settings_file

    @property
    def settings_path(self):
        """Settings json folder path."""
        return self.settings_manager.settings_path

    @property
    def xaml_file(self):
        """Wpf xaml file path."""
        return os.path.join(self.toolkit_directory, self.toolkit_name + ".xaml")

    @pyaedt_function_handler()
    def copy_xaml_template(self):
        """Copies the xaml template to local folder and rename it to be used with current application."""
        local_path = os.path.abspath(os.path.dirname(__file__))
        print("xaml file = {}".format(self.xaml_file))
        shutil.copy2(os.path.join(local_path, "wpf_template.xaml"), self.xaml_file)
        return True

    @pyaedt_function_handler()
    def _add_line_to_xml(self, line_to_add):
        with open(self.xaml_file, "r") as file:
            file = file.readlines()
        with open(self.xaml_file[:-5] + "_tmp.xaml", "w") as f:
            for line in file:
                if "</Grid>" in line:
                    f.write(line_to_add + "\n")
                f.write(line)
        shutil.move(self.xaml_file[:-5] + "_tmp.xaml", self.xaml_file)

    @pyaedt_function_handler()
    def edit_window_size(self, width=800, height=600, title="PyAEDT WPF Application", background="#FFD1CFCF"):
        """Edit the Wpf windows size.

        Parameters
        ----------
        width : int, optional
            Windows width.
        height : int, optional
            Windows height.
        title : str, optional
            Windows title.
        background : str, optional
            Windows color in hex mode.

        Returns
        -------
        bool
            `True` if succeeded.
        """
        with open(self.xaml_file, "r") as file:
            file = file.readlines()
        line_to_add = '        Title="{}" Height="{}" Width="{}" Background="{}">'.format(
            title, height, width, background
        )

        with open(self.xaml_file[:-5] + "_tmp.xaml", "w") as f:
            for line in file:
                if "       Title=" in line:
                    f.write(line_to_add + "\n")
                else:
                    f.write(line)
        shutil.move(self.xaml_file[:-5] + "_tmp.xaml", self.xaml_file)
        return True

    @pyaedt_function_handler()
    def add_label(self, name, content, x_pos, y_pos):
        """Adds a label to Wpf.

        Parameters
        ----------
        name : str
            Name of the label.
        content : str
            Content of the label.
        x_pos : float
            Horizontal position in UI.
        y_pos : float
            Vertical position in UI.

        Returns
        -------
        bool
        """
        new_label = '        <Label x:Name="{}" Content="{}" HorizontalAlignment="Left" '.format(name, content)
        new_label += 'Margin="{},{},0,0" VerticalAlignment="Top"/>'.format(x_pos, y_pos)
        self._add_line_to_xml(new_label)
        return True

    @pyaedt_function_handler()
    def add_text_box(self, name, x_pos, y_pos, width=120, callback_method=None, callback_action="LostFocus"):
        """Adds a text box to Wpf.

        Parameters
        ----------
        name : str
            Name of the text box.
        x_pos : float
            Horizontal position in UI.
        y_pos : float
            Vertical position in UI.
        width : float
            Width of the text box.
        callback_method : func
            Name of the method assigned to the call back action. `None` to disable it.
        callback_action : str, optional
            Call back action on which callback will be applied. Default is `"Click"`.

        Returns
        -------
        bool
        """
        new_label = '        <TextBox x:Name="{}" HorizontalAlignment="Left" Height="23" '.format(name)
        new_label += 'Margin="{},{},0,0" TextWrapping="Wrap" VerticalAlignment="Top" Width="{}"/>'.format(
            x_pos, y_pos, width
        )
        if callback_method:
            self._callbacks.append([name, callback_action, callback_method])
        self._add_line_to_xml(new_label)
        return True

    @pyaedt_function_handler()
    def add_combo_box(self, name, x_pos, y_pos, width=120, callback_method=None, callback_action="SelectionChanged"):
        """Adds a combo box to Wpf.

        Parameters
        ----------
        name : str
            Name of the combo.
        x_pos : float
            Horizontal position in UI.
        y_pos : float
            Vertical position in UI.
        width : float
            Width of the combo box.
        callback_method : func
            Name of the method assigned to the call back action. `None` to disable it.
        callback_action : str, optional
            Call back action on which callback will be applied. Default is `"Click"`.
        Returns
        -------
        bool
        """
        new_label = '        <ComboBox x:Name="{}" HorizontalAlignment="Left" Height="23" '.format(name)
        new_label += 'Margin="{},{},0,0" VerticalAlignment="Top" Width="{}"/>'.format(x_pos, y_pos, width)
        if callback_method:
            self._callbacks.append([name, callback_action, callback_method])
        self._add_line_to_xml(new_label)
        return True

    @pyaedt_function_handler()
    def add_check_box(self, name, content, x_pos, y_pos, callback_method=None, callback_action="Checked"):
        """Adds a check box to Wpf.

        Parameters
        ----------
        name : str
            Name of the check.
        content : str
            Caption of the check box.
        x_pos : float
            Horizontal position in UI.
        y_pos : float
            Vertical position in UI.
        callback_method : func
            Name of the method assigned to the call back action. `None` to disable it.
        callback_action : str, optional
            Call back action on which callback will be applied. Default is `"Click"`.
        Returns
        -------
        bool
        """
        new_label = '        <CheckBox x:Name="{}" Content="{}" HorizontalAlignment="Left" '.format(name, content)
        new_label += 'Margin="{},{},0,0" VerticalAlignment="Top"/>'.format(x_pos, y_pos)
        if callback_method:
            self._callbacks.append([name, callback_action, callback_method])
        self._add_line_to_xml(new_label)
        return True

    @pyaedt_function_handler()
    def add_button(self, name, content, x_pos, y_pos, width=120, callback_method=None, callback_action="Click"):
        """Adds a button to Wpf.

        Parameters
        ----------
        name : str
            Name of the button.
        content : str
            Caption of the button box.
        x_pos : float
            Horizontal position in UI.
        y_pos : float
            Vertical position in UI.
        width : float
            Button width.
        callback_method : func
            Name of the method assigned to the call back action. `None` to disable it.
        callback_action : str, optional
            Call back action on which callback will be applied. Default is `"Click"`.

        Returns
        -------
        bool
        """
        new_label = '        <Button x:Name="{}" Content="{}" HorizontalAlignment="Left" '.format(name, content)
        new_label += 'Margin="{},{},0,0" VerticalAlignment="Top" Width="{}"/>'.format(x_pos, y_pos, width)
        if callback_method:
            self._callbacks.append([name, callback_action, callback_method])
        self._add_line_to_xml(new_label)
        return True

    @pyaedt_function_handler()
    def launch_gui(self):
        """Shows the Wpf UI."""
        if sys.implementation.name == "ironpython":
            clr.AddReference("IronPython.wpf")
            import wpf

            wpf.LoadComponent(self, self.xaml_file)
            self.window = self
        else:
            stream = StreamReader(self.xaml_file)
            self.window = XamlReader.Load(stream.BaseStream)

        workarea = __import__("System.Windows").Windows.SystemParameters.WorkArea
        desktopWorkingArea = workarea
        self.window.Left = desktopWorkingArea.Right - self.window.Width
        self.window.Top = desktopWorkingArea.Top

        self.SetText = self._get_objects_from_xaml_of_type("TextBox")
        self.SetCombo = self._get_objects_from_xaml_of_type("ComboBox")
        self.SetBool = self._get_objects_from_xaml_of_type(["RadioButton", "CheckBox"])
        self.read_settings()
        uri = Uri(os.path.join(self.image_path, "pyansys-logo-black-cropped.png"))
        logo = self.get_ui_object("logo")
        logo.Source = BitmapImage(uri)
        if self._callbacks:
            for el in self._callbacks:
                self.set_callback(el[0], el[1], el[2])
            self._callbacks = []

    @pyaedt_function_handler()
    def validate_object_name_prefix(self, sender, e):
        """Validates the text box with object name prefix."""
        valid = False
        try:
            object_list = self.aedtdesign.modeler.get_matched_object_name(sender.Text + "*")
            assert object_list
            valid = True
        except AssertionError:
            pass
        self.update_textbox_status(sender, valid)

    @pyaedt_function_handler()
    def validate_string_no_spaces(self, sender, e):
        """Validates the text box with no spaces."""

        valid = False
        if sender.Text.find(" ") < 0:
            valid = True
        self.update_textbox_status(sender, valid)

    @pyaedt_function_handler()
    def validate_integer(self, sender, e):
        """Validates the text box with to integer."""

        valid = False
        try:
            value = int(sender.Text)
            valid = True
        except ValueError:
            pass
        if not valid:
            value = None
        self.update_textbox_status(sender, valid)
        return value

    @pyaedt_function_handler()
    def validate_positive_odd_integer(self, sender, e):
        """Validates the text box with to positive odd integer."""

        valid = False
        try:
            value = int(sender.Text)
            if value > 0 and value % 2 != 0:
                sender.Text = str(value)
                valid = True
        except ValueError:
            pass
        if not valid:
            value = None
        self.update_textbox_status(sender, valid)
        return value

    @pyaedt_function_handler()
    def validate_positive_integer(self, sender, e):
        """Validates the text box with to strictly positive integer."""

        valid = False
        try:
            value = int(sender.Text)
            if value > 0:
                sender.Text = str(value)
                valid = True
        except ValueError:
            pass
        if not valid:
            value = None
        self.update_textbox_status(sender, valid)
        return value

    def validate_non_negative_integer(self, sender, e):
        """Validates the text box with to non negative integer."""

        valid = False
        try:
            value = int(sender.Text)
            if value >= 0:
                sender.Text = str(value)
                valid = True
        except ValueError:
            pass
        if not valid:
            value = None
        self.update_textbox_status(sender, valid)
        return value

    @pyaedt_function_handler()
    def validate_negative_integer(self, sender, e):
        """Validates the text box with to negative integer."""
        valid = False
        try:
            value = int(sender.Text)
            if value < 0:
                valid = True
        except ValueError:
            pass
        if not valid:
            value = None
        self.update_textbox_status(sender, valid)
        return value

    @pyaedt_function_handler()
    def validate_float(self, sender, e):
        """Validates the text box with to float."""

        valid = False
        try:
            value = float(sender.Text)
            sender.Text = str(value)
            valid = True
        except ValueError:
            value = None
        self.update_textbox_status(sender, valid)
        return value

    @pyaedt_function_handler()
    def validate_float_variable(self, sender, e):
        """Validates the text box with to float variable."""
        proj_and_des_variables = self.aedtdesign.variable_manager.variable_names
        if sender.Text in proj_and_des_variables:
            self.update_textbox_status(sender, True)
        else:
            self.validate_float(sender, e)

    @pyaedt_function_handler()
    def validate_positive_float_variable(self, sender, e):
        """Validates the text box with to positive float variable."""

        proj_and_des_variables = self.aedtdesign.variable_manager.variable_names
        if sender.Text in proj_and_des_variables:
            self.update_textbox_status(sender, True)
        else:
            self.validate_positive_float(sender, e)

    @pyaedt_function_handler()
    def validate_positive_integer_variable(self, sender, e):
        """Validates the text box with to positive integer variable."""

        proj_and_des_variables = self.aedtdesign.variable_manager.variable_names
        if sender.Text in proj_and_des_variables:
            self.update_textbox_status(sender, True)
        else:
            self.validate_positive_integer(sender, e)

    @pyaedt_function_handler()
    def validate_positive_integer_global(self, sender, e):
        """Validates the text box with to positive integer global variable."""

        proj_variables = self.aedtdesign.variable_manager.project_variable_names
        if sender.Text in proj_variables:
            self.update_textbox_status(sender, True)
        else:
            self.validate_positive_integer(sender, e)

    @pyaedt_function_handler()
    def validate_positive_float_global(self, sender, e):
        """Validates the text box with to positive float global variable."""

        proj_variables = self.aedtdesign.variable_manager.project_variable_names
        if sender.Text in proj_variables:
            self.update_textbox_status(sender, True)
        else:
            self.validate_positive_float(sender, e)

    @pyaedt_function_handler()
    def validate_positive_float(self, sender, e):
        """Validates the text box with to positive float."""

        valid = False
        try:
            value = float(sender.Text)
            if value > 0:
                valid = True
        except ValueError:
            pass
        if not valid:
            value = None
        self.update_textbox_status(sender, valid)
        return value

    @pyaedt_function_handler()
    def validate_non_negative_float(self, sender, e):
        """Validates the text box with to non-negative float."""

        valid = False
        try:
            value = float(sender.Text)
            if value >= 0.0:
                valid = True
        except ValueError:
            pass
        if not valid:
            value = None
        self.update_textbox_status(sender, valid)
        return value

    @pyaedt_function_handler()
    def update_textbox_status_with_default_text(self, sender, valid, default_text):
        """Updates a text box with a default text value."""

        if not valid:
            sender.Text = default_text
            sender.BorderBrush = Brushes.Red
        else:
            sender.BorderBrush = Brushes.Green

    @pyaedt_function_handler()
    def update_textbox_status(self, sender, valid):
        """Updates a text box status."""

        if not valid:
            sender.Text = ""
            sender.BorderBrush = Brushes.Red
        else:
            sender.BorderBrush = Brushes.Green

    @pyaedt_function_handler()
    def create_child_design(self, design_name):
        """Duplicates a design and makes a link to the parent design in the settings file.

        Parameters
        ----------
        design_name : str
            Design name.

        Returns
        -------
        bool
        """
        self.aedtdesign.duplicate_design(design_name)
        self._write_parent_link()
        bool

    @pyaedt_function_handler()
    def _read_and_synch_settings_file(self):
        """Reads in existing settings data and updates the path of the library directory in case the project was
        moved to a new location, file system or operating system."""
        settings_file = self.settings_file
        if os.path.exists(settings_file):
            settings_data = self.settings_data
            with open(settings_file, "w") as f:
                settings_data["_lib_dir"] = self.aedtlib_directory
                settings_data["_toolkit_dir"] = self.toolkit_directory
                json.dump(settings_data, f, indent=4)

    @pyaedt_function_handler()
    def _write_parent_link(self):
        with open(self.local_settings_file, "w") as f:
            settings_data = {"parent": self.parent_design_name}
            json.dump(settings_data, f, indent=4)

    @pyaedt_function_handler()
    def dummy_callback(self, sender, e):
        """Dummy callback."""
        pass

    @pyaedt_function_handler()
    def display(self):
        """Displays the wpf application as a Dialoq (IronPython) or an Application (CPython)."""
        if sys.implementation.name == "ironpython":
            Window.ShowDialog(self)
        else:
            Application().Run(self.window)

    @pyaedt_function_handler()
    def open_explorer(self, sender, e):
        """Opens a windows explorer window pointing to the selected path in the sender control."""
        os_type = os.name
        if os_type == "nt":
            selected_path = os.path.normpath(sender.Text)
            if os.path.exists(selected_path):
                os_command_string = r'explorer "{}"'.format(selected_path)
                subprocess.Popen(os_command_string)
            pass

    @pyaedt_function_handler()
    def set_callback(self, control, callback, function):
        """Sets up the callback functions from the xaml GUI.

        Parameters
        ----------
        control : str
            Name of the control (Button, TextBox, etc) as a string
        callback : str
            Name of the callback function, e.g. "Click", "LostFocus", etc
        function : func
            Reference to the callback function (requires 2 arguments: sender, e)
        """
        test_control = LogicalTreeHelper.FindLogicalNode(self.window, control)
        a = getattr(test_control, callback)
        a += function

    @pyaedt_function_handler()
    def set_margin(self, object_name, margin):
        """Sets the outer dimensions of the GUI window.

        Parameters
        ----------
        object_name : str
            Object name.
        margin : list
            vector of floats [left, top, bottom, right].
        """

        myThickness = Thickness()
        myThickness.Bottom = margin[2]
        myThickness.Left = margin[0]
        myThickness.Right = margin[3]
        myThickness.Top = margin[1]
        self.get_ui_object(object_name).Margin = myThickness

    @pyaedt_function_handler()
    def assign_image(self, ui_object_name, image_file):
        """Assigns an image to an object.

        Parameters
        ----------
        ui_object_name : str
            Object name.
        image_file : str
            Image full path.

        Returns
        -------

        """
        bi = BitmapImage()
        bi.BeginInit()
        bi.CacheOption = BitmapCacheOption.OnLoad
        bi.CreateOptions = BitmapCreateOptions.IgnoreImageCache
        bi.UriSource = Uri(image_file, UriKind.RelativeOrAbsolute)
        bi.EndInit()
        self.ui[ui_object_name].Source = bi

    @pyaedt_function_handler()
    def set_visible(self, object_list):
        """Defines one or more GUI objects to be visible.

        Parameters
        ----------
        object_list : list
            List of objects to make visible.

        Returns
        -------

        """
        if isinstance(object_list, str):
            object_list = [object_list]
        for object_name in object_list:
            self.get_ui_object(object_name).Visibility = Visibility.Visible

    @pyaedt_function_handler()
    def set_hidden(self, object_list):
        """Defines one or more GUI objects to be hidden.

        Parameters
        ----------
        object_list : list
            List of objects to make hidden.

        Returns
        -------

        """
        if isinstance(object_list, str):
            object_list = [object_list]
        for object_name in object_list:
            self.get_ui_object(object_name).Visibility = Visibility.Hidden

    @pyaedt_function_handler()
    def wait_cursor(self):
        """Turns on the "Wait" cursor and stores the current cursor"""
        self.previous_cursor = self.Cursor
        self.Cursor = Input.Cursors.Wait

    @pyaedt_function_handler()
    def standard_cursor(self):
        """Restores the current cursor"""
        self.Cursor = self.previous_cursor

    @pyaedt_function_handler()
    def _get_objects_from_xaml_of_type(self, type_list):
        if isinstance(type_list, str):
            type_list = [type_list]
        text_list = []
        with open(self.xaml_file, "r") as f:
            for line in f:
                rstrip_line = line.lstrip()[1:]
                pp = rstrip_line.find(" ")
                object_type = rstrip_line[0:pp]
                if object_type in type_list:
                    attribute_list = rstrip_line.split(" ")
                    name = attribute_list[1].split('"')[1]
                    text_list.append(name)
        return text_list

    @pyaedt_function_handler()
    def message_box(self, text, caption=None, buttons=None, icon=None):
        """Message Box."""
        return message_box(text, caption, buttons, icon)

    @pyaedt_function_handler()
    def ok_cancel_message_box(self, text, caption=None, icon=None):
        response = message_box(text, caption, "OKCancel", icon)
        if response == DialogResult.OK:
            return True
        else:
            return False

    @pyaedt_function_handler()
    def add_combo_items(self, combo_box_name, options, default=None):
        """Fills a combo box with a list of options and sets the selected value if nothing is present already.

        Parameters
        ----------
        combo_box_name : str
            Object name.
        options : list
            list of options.
        default : str
            Default value.

        Returns
        -------

        """
        if not isinstance(options, list):
            options = [options]
        control = self.get_ui_object(combo_box_name)
        for each_option in options:
            control.Items.Add(each_option)
        if default and not control.SelectedValue:
            control.SelectedValue = default

    @pyaedt_function_handler()
    def get_combobox_selection(self, ui_object_name):
        """Returns the selected value from a combobox
        Parameters
        ----------
        ui_object_name : str
            Object name.

        Returns str
            The selected value.
        """
        control = self.get_ui_object(ui_object_name)
        item_selected = str(control.SelectedItem)
        return item_selected

    @pyaedt_function_handler()
    def clear_combobox_items(self, ui_object_name):
        """ """
        control = self.get_ui_object(ui_object_name)
        control.Items.Clear()

    @pyaedt_function_handler()
    def set_text_value(self, ui_object_name, text_val):
        """Sets a text box value.

        Parameters
        ----------
        ui_object_name : str
            Object name.
        text_val : str
            Text value to apply.
        """
        control = self.get_ui_object(ui_object_name)
        control.Text = text_val

    @pyaedt_function_handler()
    def get_text_value(self, ui_object_name):
        """ """
        control = self.get_ui_object(ui_object_name)
        return control.Text

    @pyaedt_function_handler()
    def get_checkbox_status(self, ui_object_name):
        """ """
        control = self.get_ui_object(ui_object_name)
        if control.IsChecked:
            return True
        else:
            return False

    @pyaedt_function_handler()
    def set_chechbox_status(self, ui_object_name, flag=True):
        """Sets a check box value.

        Parameters
        ----------
        ui_object_name : str
            Object name.
        flag : bool
            Check Box status.
        """
        control = self.get_ui_object(ui_object_name)
        control.IsChecked = flag

    @pyaedt_function_handler()
    def get_ui_object(self, control_name):
        """Gets a UI object.

        Parameters
        ----------
        control_name : str
            Object name.
        """
        wpf_control = LogicalTreeHelper.FindLogicalNode(self.window, control_name)
        assert wpf_control, "WPF GUI object name {0} does not exist !".format(control_name)
        return wpf_control

    @pyaedt_function_handler()
    def read_settings(self):
        """Reads the setting data from the toolkit settings file in the parent design

        Returns
        -------
        dict
            Dictionary of settings data
        """
        settings_data = self.settings_data

        if settings_data:
            if len(settings_data) > 1:
                for text_control in self.SetText:
                    wpf_control = self.get_ui_object(text_control)
                    try:
                        txt_line = settings_data[text_control]
                        wpf_control.Text = txt_line
                        if txt_line:
                            wpf_control.BorderBrush = Brushes.Green
                    except KeyError:
                        if wpf_control.IsEnabled:
                            wpf_control.BorderBrush = Brushes.Red

                for combo_control in self.SetCombo:
                    wpf_control = self.get_ui_object(combo_control)
                    try:
                        txt_line = settings_data[wpf_control.Name]
                        wpf_control.SelectedValue = txt_line
                        try:
                            self.aedtdesign.logger.info("Trying to set: " + txt_line)
                        except:
                            pass
                        if txt_line:
                            wpf_control.BorderBrush = Brushes.Green
                    except KeyError:
                        if wpf_control.IsEnabled:
                            wpf_control.BorderBrush = Brushes.Red

                for bool_control in self.SetBool:
                    wpf_control = self.get_ui_object(bool_control)
                    try:
                        txt_line = int(settings_data[wpf_control.Name])
                    except KeyError:
                        txt_line = 0
                    except ValueError:
                        txt_line = 0
                    wpf_control.IsChecked = bool(txt_line)
                    if txt_line:
                        wpf_control.BorderBrush = Brushes.Green

        return settings_data

    @pyaedt_function_handler()
    def write_settings(self, user_defined_data=None):
        """Writes UI settings Textbox, Checkbox, Combobox only at present
            also write any user defined data from a json-serializable dictionary

        Parameters
        ----------
        user_defined_data: dict
            Dictionary with arbitrary user data (needs to be json serializable).
        """
        settings_data = self.settings_data
        if not settings_data:
            return False
        with open(self.settings_file, "w") as f:

            for text_control in self.SetText:
                wpf_control = self.get_ui_object(text_control)
                settings_data[wpf_control.Name] = wpf_control.Text

            for bool_control in self.SetBool:
                wpf_control = self.get_ui_object(bool_control)
                if wpf_control.IsChecked:
                    settings_data[wpf_control.Name] = "1"
                else:
                    settings_data[wpf_control.Name] = "0"

            for combo_control in self.SetCombo:
                wpf_control = self.get_ui_object(combo_control)
                settings_data[wpf_control.Name] = wpf_control.Text

            if isinstance(user_defined_data, dict):
                for key, entry in user_defined_data.items():
                    settings_data[key] = entry

            json.dump(settings_data, f, indent=4)

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):

        # write the UI data to json
        self.write_settings()

        # If an error occurred in the context of this object, close the window to avoid AEDT crashing
        if ex_type and self.window:
            self.Close()



"""
This module contains the `PostProcessor` class.

It contains all advanced postprocessing functionalities that require Python 3.x packages like NumPy and Matplotlib.
"""
from __future__ import absolute_import  # noreorder

import os
import time
import warnings

from pyaedt.generic.general_methods import is_ironpython
from pyaedt.generic.general_methods import pyaedt_function_handler
from pyaedt.generic.plot import ModelPlotter
from pyaedt.modules.PostProcessor import PostProcessor as Post

if not is_ironpython:
    try:
        import numpy as np
    except ImportError:
        warnings.warn(
            "The NumPy module is required to run some functionalities of PostProcess.\n"
            "Install with \n\npip install numpy\n\nRequires CPython."
        )

    try:
        from IPython.display import Image

        ipython_available = True
    except ImportError:
        warnings.warn(
            "The Ipython module is required to run some functionalities of PostProcess.\n"
            "Install with \n\npip install ipython\n\nRequires CPython."
        )


class PostProcessor(Post):
    """Contains advanced postprocessing functionalities that require Python 3.x packages like NumPy and Matplotlib.

    Parameters
    ----------
    app :
        Inherited parent object.

    Examples
    --------
    Basic usage demonstrated with an HFSS, Maxwell, or any other design:

    >>> from pyaedt import Hfss
    >>> aedtapp = Hfss()
    >>> post = aedtapp.post
    """

    def __init__(self, app):
        Post.__init__(self, app)

    @pyaedt_function_handler()
    def nb_display(self, show_axis=True, show_grid=True, show_ruler=True):
        """Show the Jupyter Notebook display.

          .. note::
              .assign_curvature_extraction Jupyter Notebook is not supported by IronPython.

         Parameters
         ----------
         show_axis : bool, optional
             Whether to show the axes. The default is ``True``.
         show_grid : bool, optional
             Whether to show the grid. The default is ``True``.
         show_ruler : bool, optional
             Whether to show the ruler. The default is ``True``.

        Returns
        -------
        :class:`IPython.core.display.Image`
            Jupyter notebook image.

        """
        file_name = self.export_model_picture(show_axis=show_axis, show_grid=show_grid, show_ruler=show_ruler)
        return Image(file_name, width=500)

    @pyaedt_function_handler()
    def get_efields_data(self, setup_sweep_name="", ff_setup="Infinite Sphere1", freq="All"):
        """Compute Etheta and EPhi.

        .. warning::
           This method requires NumPy to be installed on your machine.


        Parameters
        ----------
        setup_sweep_name : str, optional
            Name of the setup for computing the report. The default is ``""``, in
            which case the nominal adaptive is applied.
        ff_setup : str, optional
            Far field setup. The default is ``"Infinite Sphere1"``.
        freq : str, optional
            The default is ``"All"``.

        Returns
        -------
        np.ndarray
            numpy array containing ``[theta_range, phi_range, Etheta, Ephi]``.
        """
        if not setup_sweep_name:
            setup_sweep_name = self._app.nominal_adaptive
        results_dict = {}
        all_sources = self.post_osolution.GetAllSources()
        # assuming only 1 mode
        all_sources_with_modes = [s + ":1" for s in all_sources]

        for n, source in enumerate(all_sources_with_modes):
            edit_sources_ctxt = [["IncludePortPostProcessing:=", False, "SpecifySystemPower:=", False]]
            for m, each in enumerate(all_sources_with_modes):
                if n == m:  # set only 1 source to 1W, all the rest to 0
                    mag = 1
                else:
                    mag = 0
                phase = 0
                edit_sources_ctxt.append(
                    ["Name:=", "{}".format(each), "Magnitude:=", "{}W".format(mag), "Phase:=", "{}deg".format(phase)]
                )
            self.post_osolution.EditSources(edit_sources_ctxt)

            ctxt = ["Context:=", ff_setup]

            sweeps = ["Theta:=", ["All"], "Phi:=", ["All"], "Freq:=", [freq]]

            trace_name = "rETheta"
            solnData = self.get_far_field_data(
                setup_sweep_name=setup_sweep_name, domain=ff_setup, expression=trace_name
            )

            data = solnData.nominal_variation

            theta_vals = np.degrees(np.array(data.GetSweepValues("Theta")))
            phi_vals = np.degrees(np.array(data.GetSweepValues("Phi")))
            # phi is outer loop
            theta_unique = np.unique(theta_vals)
            phi_unique = np.unique(phi_vals)
            theta_range = np.linspace(np.min(theta_vals), np.max(theta_vals), np.size(theta_unique))
            phi_range = np.linspace(np.min(phi_vals), np.max(phi_vals), np.size(phi_unique))
            real_theta = np.array(data.GetRealDataValues(trace_name))
            imag_theta = np.array(data.GetImagDataValues(trace_name))

            trace_name = "rEPhi"
            solnData = self.get_far_field_data(
                setup_sweep_name=setup_sweep_name, domain=ff_setup, expression=trace_name
            )
            data = solnData.nominal_variation

            real_phi = np.array(data.GetRealDataValues(trace_name))
            imag_phi = np.array(data.GetImagDataValues(trace_name))

            Etheta = np.vectorize(complex)(real_theta, imag_theta)
            Ephi = np.vectorize(complex)(real_phi, imag_phi)
            source_name_without_mode = source.replace(":1", "")
            results_dict[source_name_without_mode] = [theta_range, phi_range, Etheta, Ephi]
        return results_dict

    @pyaedt_function_handler()
    def get_model_plotter_geometries(
        self,
        objects=None,
        plot_as_separate_objects=True,
        plot_air_objects=False,
        force_opacity_value=None,
        array_coordinates=None,
        generate_mesh=True,
    ):
        """Initialize the Model Plotter object with actual modeler objects and return it.

         Parameters
         ----------
         objects : list, optional
             Optional list of objects to plot. If `None` all objects will be exported.
         plot_as_separate_objects : bool, optional
             Plot each object separately. It may require more time to export from AEDT.
         plot_air_objects : bool, optional
             Plot also air and vacuum objects.
         force_opacity_value : float, optional
             Opacity value between 0 and 1 to be applied to all model.
             If `None` aedt opacity will be applied to each object.
        array_coordinates : list of list
            List of array element centers. The modeler objects will be duplicated and translated.
            List of [[x1,y1,z1], [x2,y2,z2]...].

         Returns
         -------
         :class:`pyaedt.generic.plot.ModelPlotter`
             Model Object.
        """
        assert self._app._aedt_version >= "2021.2", self.logger.error("Object is supported from AEDT 2021 R2.")
        files = self.export_model_obj(
            obj_list=objects,
            export_as_single_objects=plot_as_separate_objects,
            air_objects=plot_air_objects,
        )
        if not files:
            self.logger.warning("No Objects exported. Try other options or include Air objects.")
            return False

        model = ModelPlotter()

        for file in files:
            if force_opacity_value:
                model.add_object(file[0], file[1], force_opacity_value, self.modeler.model_units)
            else:
                model.add_object(file[0], file[1], file[2], self.modeler.model_units)
        model.array_coordinates = array_coordinates
        if generate_mesh:
            model.generate_geometry_mesh()
        return model

    @pyaedt_function_handler()
    def plot_model_obj(
        self,
        objects=None,
        show=True,
        export_path=None,
        plot_as_separate_objects=True,
        plot_air_objects=False,
        force_opacity_value=None,
        clean_files=False,
        array_coordinates=None,
    ):
        """Plot the model or a substet of objects.

        Parameters
        ----------
        objects : list, optional
            Optional list of objects to plot. If `None` all objects will be exported.
        show : bool, optional
            Show the plot after generation or simply return the
            generated Class for more customization before plot.
        export_path : str, optional
            If available, an image is saved to file. If `None` no image will be saved.
        plot_as_separate_objects : bool, optional
            Plot each object separately. It may require more time to export from AEDT.
        plot_air_objects : bool, optional
            Plot also air and vacuum objects.
        force_opacity_value : float, optional
            Opacity value between 0 and 1 to be applied to all model.
            If `None` aedt opacity will be applied to each object.
        clean_files : bool, optional
            Clean created files after plot. Cache is mainteined into the model object returned.
        array_coordinates : list of list
            List of array element centers. The modeler objects will be duplicated and translated.
            List of [[x1,y1,z1], [x2,y2,z2]...].

        Returns
        -------
        :class:`pyaedt.generic.plot.ModelPlotter`
            Model Object.
        """
        model = self.get_model_plotter_geometries(
            objects=objects,
            plot_as_separate_objects=plot_as_separate_objects,
            plot_air_objects=plot_air_objects,
            force_opacity_value=force_opacity_value,
            array_coordinates=array_coordinates,
            generate_mesh=False,
        )

        if not show:
            model.off_screen = True
        if export_path:
            model.plot(export_path)
        elif show:
            model.plot()
        if clean_files:
            model.clean_cache_and_files(clean_cache=False)
        return model

    @pyaedt_function_handler()
    def plot_field_from_fieldplot(
        self,
        plotname,
        project_path="",
        meshplot=False,
        imageformat="jpg",
        view="isometric",
        plot_label="Temperature",
        plot_folder=None,
        show=True,
        scale_min=None,
        scale_max=None,
    ):
        """Export a field plot to an image file (JPG or PNG) using Python PyVista.

        .. note::
           The PyVista module rebuilds the mesh and the overlap fields on the mesh.

        Parameters
        ----------
        plotname : str
            Name of the field plot to export.
        project_path : str, optional
            Path for saving the image file. The default is ``""``.
        meshplot : bool, optional
            Whether to create and plot the mesh over the fields. The
            default is ``False``.
        imageformat : str, optional
            Format of the image file. Options are ``"jpg"``,
            ``"png"``, ``"svg"``, and ``"webp"``. The default is
            ``"jpg"``.
        view : str, optional
            View to export. Options are ``isometric``, ``top``, ``front``,
             ``left``, ``all``.. The default is ``"iso"``. If ``"all"``, all views are exported.
        plot_label : str, optional
            Type of the plot. The default is ``"Temperature"``.
        plot_folder : str, optional
            Plot folder to update before exporting the field.
            The default is ``None``, in which case all plot
            folders are updated.
        show : bool, optional
            Export Image without plotting on UI.
        scale_min : float, optional
            Fix the Scale Minimum value.
        scale_max : float, optional
            Fix the Scale Maximum value.

        Returns
        -------
        :class:`pyaedt.generic.plot.ModelPlotter`
            Model Object.
        """
        if not plot_folder:
            self.ofieldsreporter.UpdateAllFieldsPlots()
        else:
            self.ofieldsreporter.UpdateQuantityFieldsPlots(plot_folder)

        start = time.time()
        file_to_add = self.export_field_plot(plotname, self._app.working_directory)

        model = self.get_model_plotter_geometries(generate_mesh=False)

        model.off_screen = not show
        if file_to_add:
            model.add_field_from_file(file_to_add, coordinate_units=self.modeler.model_units, show_edges=meshplot)
            if plot_label:
                model.fields[0].label = plot_label

        model.view = view

        if scale_min and scale_max:
            model.range_min = scale_min
            model.range_max = scale_max
        if show or project_path:
            model.plot(os.path.join(project_path, self._app.project_name + "." + imageformat))
            model.clean_cache_and_files(clean_cache=False)
        return model

    @pyaedt_function_handler()
    def animate_fields_from_aedtplt(
        self,
        plotname,
        plot_folder=None,
        meshplot=False,
        variation_variable="Phi",
        variation_list=["0deg"],
        project_path="",
        export_gif=False,
        show=True,
    ):
        """Generate a field plot to an image file (JPG or PNG) using PyVista.

        .. note::
           The PyVista module rebuilds the mesh and the overlap fields on the mesh.

        Parameters
        ----------
        plotname : str
            Name of the plot or the name of the object.
        plot_folder : str, optional
            Name of the folder in which the plot resides. The default
            is ``None``.
        variation_variable : str, optional
            Variable to vary. The default is ``"Phi"``.
        variation_list : list, optional
            List of variation values with units. The default is
            ``["0deg"]``.
        project_path : str, optional
            Path for the export. The default is ``""`` which export file in working_directory.
        meshplot : bool, optional
             The default is ``False``. Valid from Version 2021.2.
        export_gif : bool, optional
             The default is ``False``.
                show=False,
        show : bool, optional
             Generate the animation without showing an interactive plot.  The default is ``True``.

        Returns
        -------
        :class:`pyaedt.generic.plot.ModelPlotter`
            Model Object.
        """
        if not plot_folder:
            self.ofieldsreporter.UpdateAllFieldsPlots()
        else:
            self.ofieldsreporter.UpdateQuantityFieldsPlots(plot_folder)

        fields_to_add = []
        if not project_path:
            project_path = self._app.working_directory
        for el in variation_list:
            self._app._odesign.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:FieldsPostProcessorTab",
                        ["NAME:PropServers", "FieldsReporter:" + plotname],
                        ["NAME:ChangedProps", ["NAME:" + variation_variable, "Value:=", el]],
                    ],
                ]
            )
            fields_to_add.append(
                self.export_field_plot(plotname, project_path, plotname + variation_variable + str(el))
            )

        model = self.get_model_plotter_geometries(generate_mesh=False)
        model.off_screen = not show

        if fields_to_add:
            model.add_frames_from_file(fields_to_add)
        if export_gif:
            model.gif_file = os.path.join(self._app.working_directory, self._app.project_name + ".gif")

        if show or export_gif:
            model.animate()
            model.clean_cache_and_files(clean_cache=False)
        return model

    @pyaedt_function_handler()
    def animate_fields_from_aedtplt_2(
        self,
        quantityname,
        object_list,
        plottype,
        meshplot=False,
        setup_name=None,
        intrinsic_dict={},
        variation_variable="Phi",
        variation_list=["0deg"],
        project_path="",
        export_gif=False,
        show=True,
        zoom=None,
    ):
        """Generate a field plot to an animated gif file using PyVista.

         .. note::
            The PyVista module rebuilds the mesh and the overlap fields on the mesh.

        This method creates the plot and exports it.
        It is an alternative to the method :func:`animate_fields_from_aedtplt`,
        which uses an existing plot.

        Parameters
        ----------
        quantityname : str
            Name of the plot or the name of the object.
        object_list : list, optional
            Name of the ``folderplot`` folder.
        plottype : str
            Type of the plot. Options are ``"Surface"``, ``"Volume"``, and
            ``"CutPlane"``.
        meshplot : bool, optional
            The default is ``False``.
        setup_name : str, optional
            Name of the setup (sweep) to use for the export. The default is
            ``None``.
        intrinsic_dict : dict, optional
            Intrinsic dictionary that is needed for the export.
            The default is ``{}``.
        variation_variable : str, optional
            Variable to vary. The default is ``"Phi"``.
        variation_list : list, option
            List of variation values with units. The default is
            ``["0deg"]``.
        project_path : str, optional
            Path for the export. The default is ``""`` which export file in working_directory.
        export_gif : bool, optional
            Whether to export to a GIF file. The default is ``False``,
            in which case the plot is exported to a JPG file.
        show : bool, optional
            Generate the animation without showing an interactive plot.  The default is ``True``.
        zoom : float, optional
            Zoom factor.

        Returns
        -------
        :class:`pyaedt.generic.plot.ModelPlotter`
            Model Object.
        """
        if not project_path:
            project_path = self._app.working_directory

        v = 0
        fields_to_add = []
        for el in variation_list:
            intrinsic_dict[variation_variable] = el
            if plottype == "Surface":
                plotf = self.create_fieldplot_surface(object_list, quantityname, setup_name, intrinsic_dict)
            elif plottype == "Volume":
                plotf = self.create_fieldplot_volume(object_list, quantityname, setup_name, intrinsic_dict)
            else:
                plotf = self.create_fieldplot_cutplane(object_list, quantityname, setup_name, intrinsic_dict)
            if plotf:
                file_to_add = self.export_field_plot(plotf.name, project_path, plotf.name + str(v))
                if file_to_add:
                    fields_to_add.append(file_to_add)
                plotf.delete()
            v += 1
        model = self.get_model_plotter_geometries(generate_mesh=False)
        model.off_screen = not show

        if fields_to_add:
            model.add_frames_from_file(fields_to_add)
        if export_gif:
            model.gif_file = os.path.join(self._app.working_directory, self._app.project_name + ".gif")
        if zoom:
            model.zoom = zoom
        if show or export_gif:
            model.animate()
            model.clean_cache_and_files(clean_cache=False)

        return model

    @pyaedt_function_handler()
    def create_3d_plot(
        self, solution_data, nominal_sweep=None, nominal_value=None, primary_sweep="Theta", secondary_sweep="Phi"
    ):
        """Create a 3D plot using Matplotlib.

        Parameters
        ----------
        solution_data : :class:`pyaedt.modules.solutions.SolutionData`
            Input data for the solution.
        nominal_sweep : str, optional
            Name of the nominal sweep. The default is ``None``.
        nominal_value : str, optional
            Value for the nominal sweep. The default is ``None``.
        primary_sweep : str, optional
            Primary sweep. The default is ``"Theta"``.
        secondary_sweep : str, optional
            Secondary sweep. The default is ``"Phi"``.

        Returns
        -------
         bool
             ``True`` when successful, ``False`` when failed.
        """
        if nominal_value:
            solution_data.intrinsics[nominal_sweep] = nominal_value
        if nominal_value:
            solution_data.primary_sweep = primary_sweep
        return solution_data.plot_3d(x_axis=primary_sweep, y_axis=secondary_sweep)

    @pyaedt_function_handler()
    def plot_scene(self, frames_list, output_gif_path, norm_index=0, dy_rng=0, fps=30, show=True):
        """Plot the current model 3D scene with overlapping animation coming from a file list and save the gif.


        Parameters
        ----------
        frames_list : list or str
            File list containing animation frames to plot in csv format or
            path to a txt index file containing full path to csv files.
        output_gif_path : str
            Full path to output gif file.
        norm_index : int, optional
            Pick the frame to use to normalize your images.
            Data is already saved as dB : 100 for usual traffic scenes.
        dy_rng : int, optional
            Specify how many dB below you would like to specify the range_min.
            Tweak this a couple of times with small number of frames.
        fps : int, optional
            Frames per Second.
        show : bool, optional
            Either if show or only export gif.

        Returns
        -------

        """
        if isinstance(frames_list, str) and os.path.exists(frames_list):
            with open(frames_list, "r") as f:
                lines = f.read()
                temp_list = lines.splitlines()
            frames_paths_list = [i for i in temp_list]
        elif isinstance(frames_list, str):
            self.logger.error("Path doesn't exists")
            return False
        else:
            frames_paths_list = frames_list
        scene = self.get_model_plotter_geometries(generate_mesh=False)

        norm_data = np.loadtxt(frames_paths_list[norm_index], skiprows=1, delimiter=",")
        norm_val = norm_data[:, -1]
        v_max = np.max(norm_val)
        v_min = v_max - dy_rng
        scene.add_frames_from_file(frames_paths_list, log_scale=False, color_map="jet", header_lines=1, opacity=0.8)

        # Specifying the attributes of the scene through the ModelPlotter object
        scene.off_screen = not show
        scene.isometric_view = False
        scene.range_min = v_min
        scene.range_max = v_max
        scene.show_grid = False
        scene.windows_size = [1920, 1080]
        scene.show_legend = False
        scene.show_bounding_box = False
        scene.legend = False
        scene.frame_per_seconds = fps
        scene.camera_position = "yz"
        scene.zoom = 2
        scene.bounding_box = False
        scene.color_bar = False
        scene.gif_file = output_gif_path  # This gif may be a bit slower so we can speed it up a bit
        scene.animate()



import os
from collections import OrderedDict

try:
    import joblib
except ImportError:
    pass
try:
    import numpy as np
except ImportError:
    pass

from pyaedt import constants
from pyaedt.generic.general_methods import generate_unique_name
from pyaedt.generic.general_methods import pyaedt_function_handler
from pyaedt.modules.MaterialLib import Material
from pyaedt.generic.general_methods import is_ironpython

LAYERS = {"s": "signal", "g": "ground", "d": "dielectric"}


def _replace_by_underscore(character, string):
    if not isinstance(character, str):
        raise TypeError("character must be str")
    if not isinstance(character, str):
        raise TypeError("string must be str")
    reformat_name = list(string)
    while character in reformat_name:
        index = reformat_name.index(character)
        reformat_name[index] = "_"
    return "".join(reformat_name)


class NamedVariable(object):
    """Cast PyAEDT variable object to simplify getters and setters in Stackup3D.

    Parameters
    ----------
    application : :class:`pyaedt.hfss.Hfss
        HFSS design or project where the variable is to be created.
    name : str
        The name of the variable. If the the name begins with an '$', the variable will be a project variable.
        Otherwise, it will be a design variable.
    expression : str
        Expression of the value.

    Examples
    --------

    >>> from pyaedt import Hfss
    >>> from pyaedt.modeler.stackup_3d import Stackup3D
    >>> hfss = Hfss()
    >>> my_frequency = NamedVariable(hfss, "my_frequency", "900000Hz")
    >>> wave_length_formula = "c0/" + my_frequency.name
    >>> my_wave_length = NamedVariable(hfss, "my_wave_length", wave_length_formula)
    >>> my_permittivity = NamedVariable(hfss, "my_permittivity", "2.2")
    >>> my_wave_length.expression = my_wave_length.expression + "/" + my_permittivity.name

    """

    def __init__(self, application, name, expression):
        self._application = application
        self._name = name
        self._expression = expression
        application[name] = expression

    @property
    def _variable(self):
        return self._application.variable_manager.variables[self._name]

    @property
    def name(self):
        """Name of the variable as a string."""
        return self._name

    @property
    def expression(self):
        """Expression of the variable as a string."""
        return self._expression

    @expression.setter
    def expression(self, expression):
        """Set the expression of the variable.

        Parameters
        ----------
        expression: str
            Value expression of the variable."""
        if isinstance(expression, str):
            self._expression = expression
            self._application[self.name] = expression
        else:
            self._application.logger.error("Expression must be a string")

    @property
    def unit_system(self):
        """Unit system of the expression as a string."""
        return self._variable.unit_system

    @property
    def units(self):
        """Units."""
        return self._variable.units

    @property
    def value(self):
        """Value."""
        return self._variable.value

    @property
    def numeric_value(self):
        """Numeric part of the expression as a float value."""
        return self._variable.numeric_value

    @property
    def evaluated_value(self):
        """String that combines the numeric value and the units."""
        return self._variable.evaluated_value

    @pyaedt_function_handler()
    def hide_variable(self, value=True):
        """Set the variable to a hidden variable.

        Parameters
        ----------
        value : bool, optional
            Whether the variable is a hidden variable. The default is ``True``.

        Returns
        bool
        """
        self._application.variable_manager[self._name].hidden = value
        return True

    @pyaedt_function_handler()
    def read_only_variable(self, value=True):
        """Set the variable to a read-only variable.

        Parameters
        ----------
        value : bool, optional
            Whether the variable is a read-only variable. The default is ``True``.

        Returns
        -------
        bool
        """
        self._application.variable_manager[self._name].read_only = value
        return True


class Layer3D(object):
    """Provides a class for a management of a parametric layer in 3D Modeler."""

    def __init__(
        self,
        stackup,
        app,
        name,
        layer_type="S",
        material="copper",
        thickness=0.035,
        fill_material="FR4_epoxy",
        index=1,
    ):
        self._stackup = stackup
        self._index = index
        self._app = app
        self._name = name
        layer_position = "layer_" + name + "_position"
        self._position = NamedVariable(app, layer_position, "0mm")
        self._thickness = None
        self._layer_type = LAYERS.get(layer_type.lower())

        self._obj_3d = []
        obj_3d = None
        self._material = self.duplicate_parametrize_material(material)
        self._material_name = self._material.name
        if self._layer_type != "dielectric":
            self._fill_material = self.duplicate_parametrize_material(fill_material)
            self._fill_material_name = self._fill_material.name
        self._thickness_variable = self._name + "_thickness"
        if thickness:
            self._thickness = NamedVariable(self._app, self._thickness_variable, str(thickness) + "mm")
        if self._layer_type == "dielectric":
            obj_3d = self._app.modeler.create_box(
                ["dielectric_x_position", "dielectric_y_position", layer_position],
                ["dielectric_length", "dielectric_width", self._thickness_variable],
                name=self._name,
                matname=self._material_name,
            )
        elif self._layer_type == "ground":
            if thickness:
                obj_3d = self._app.modeler.create_box(
                    ["dielectric_x_position", "dielectric_y_position", layer_position],
                    ["dielectric_length", "dielectric_width", self._thickness_variable],
                    name=self._name,
                    matname=self._material_name,
                )

            else:
                obj_3d = self._app.modeler.create_rectangle(
                    "Z",
                    ["dielectric_x_position", "dielectric_y_position", layer_position],
                    ["dielectric_length", "dielectric_width"],
                    name=self._name,
                    matname=self._material_name,
                )
        elif self._layer_type == "signal":
            if thickness:
                obj_3d = self._app.modeler.create_box(
                    ["dielectric_x_position", "dielectric_y_position", layer_position],
                    ["dielectric_length", "dielectric_width", self._thickness_variable],
                    name=self._name,
                    matname=self._fill_material,
                )
            else:
                obj_3d = self._app.modeler.create_rectangle(
                    "Z",
                    ["dielectric_x_position", "dielectric_y_position", layer_position],
                    ["dielectric_length", "dielectric_width"],
                    name=self._name,
                    matname=self._fill_material,
                )
        obj_3d.group_name = "Layer_{}".format(self._name)
        if obj_3d:
            self._obj_3d.append(obj_3d)
        else:
            self._app.logger.error("Generation of the ground layer does not work.")

    @property
    def name(self):
        """Layer name.

        Returns
        -------
        str
        """
        return self._name

    @property
    def number(self):
        """Layer ID.

        Returns
        -------
        int
        """
        return self._index

    @property
    def material_name(self):
        """Material name.

        Returns
        -------
        str
        """
        return self._material_name

    @property
    def material(self):
        """Material.

        Returns
        -------
        :class:`pyaedt.modules.Material.Material`
            Material.
        """
        return self._material

    @property
    def filling_material(self):
        """Fill material.

        Returns
        -------
        :class:`pyaedt.modules.Material.Material`
            Material.
        """
        return self._fill_material

    @property
    def filling_material_name(self):
        """Fill material name.

        Returns
        -------
        str
        """
        return self._fill_material_name

    @property
    def thickness(self):
        """Thickness variable.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._thickness

    @property
    def thickness_value(self):
        """Thickness value.


        Returns
        -------
        float, str
        """
        return self._thickness.value

    @thickness.setter
    def thickness(self, value):
        self._thickness.expression = value

    @property
    def elevation(self):
        """Layer elevation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._position

    @property
    def elevation_value(self):
        """Layer elevation value.

        Returns
        -------
        str, float
        """
        return self._app.variable_manager[self._position.name].value

    @pyaedt_function_handler()
    def duplicate_parametrize_material(self, material_name, cloned_material_name=None, list_of_properties=None):
        """Duplicate a material and parametrize all properties.

        Parameters
        ----------
        material_name : str
            Name of origin material
        cloned_material_name : str, optional
            Name of destination material. The default is ``None``.
        list_of_properties : list, optional
            Properties to parametrize. The default is ``None``.

        Returns
        -------
        :class:`pyaedt.modules.Material.Material`
            Material object.
        """
        application = self._app
        if isinstance(material_name, Material):
            return material_name
        if isinstance(cloned_material_name, Material):
            return cloned_material_name
        if self._app.materials.checkifmaterialexists(material_name):
            if not cloned_material_name:
                cloned_material_name = "cloned_" + material_name
            if not self._app.materials.checkifmaterialexists(cloned_material_name):
                if not list_of_properties:
                    cloned_material = application.materials.duplicate_material(material_name, cloned_material_name)
                    permittivity = cloned_material.permittivity.value
                    permeability = cloned_material.permeability.value
                    conductivity = cloned_material.conductivity.value
                    dielectric_loss_tan = cloned_material.dielectric_loss_tangent.value
                    magnetic_loss_tan = cloned_material.magnetic_loss_tangent.value
                    reformat_name = _replace_by_underscore(" ", cloned_material_name)
                    reformat_name = _replace_by_underscore("(", reformat_name)
                    reformat_name = _replace_by_underscore(")", reformat_name)
                    reformat_name = _replace_by_underscore("/", reformat_name)
                    reformat_name = _replace_by_underscore("-", reformat_name)
                    reformat_name = _replace_by_underscore(".", reformat_name)
                    reformat_name = _replace_by_underscore(",", reformat_name)
                    permittivity_variable = "$" + reformat_name + "_permittivity"
                    permeability_variable = "$" + reformat_name + "_permeability"
                    conductivity_variable = "$" + reformat_name + "_conductivity"
                    dielectric_loss_variable = "$" + reformat_name + "_dielectric_loss"
                    magnetic_loss_variable = "$" + reformat_name + "_magnetic_loss"
                    application[permittivity_variable] = str(permittivity)
                    application[permeability_variable] = str(permeability)
                    application[conductivity_variable] = str(conductivity)
                    application[dielectric_loss_variable] = str(dielectric_loss_tan)
                    application[magnetic_loss_variable] = str(magnetic_loss_tan)
                    cloned_material.permittivity = permittivity_variable
                    cloned_material.permeability = permeability_variable
                    cloned_material.conductivity = conductivity_variable
                    cloned_material.dielectric_loss_tangent = dielectric_loss_variable
                    cloned_material.magnetic_loss_tangent = magnetic_loss_variable
                    return cloned_material
            else:
                return application.materials[cloned_material_name]
        else:
            application.logger.error("The material name %s doesn't exist" % material_name)
            return None

    @pyaedt_function_handler()
    def add_patch(
        self,
        frequency,
        patch_width,
        patch_length=None,
        patch_position_x=0,
        patch_position_y=0,
        patch_name=None,
        axis="X",
    ):
        """Create a parametric patch.

        Parameters
        ----------
        frequency : float
            Frequency value for the patch calculation in Hz.
        patch_width : float
            Patch width.
        patch_length : float, optional
            Patch length. The default is ``None``.
        patch_position_x : float, optional
            Patch start x position.
        patch_position_y : float, optional
            Patch start y position. The default is ``0.``
        patch_name : str, optional
            Patch name. The default is ``None``.
        axis : str, optional
            Line orientation axis. The default is ``"X"``.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Patch`
        """
        if not patch_name:
            patch_name = generate_unique_name("{}_patch".format(self._name), n=3)
        lst = self._stackup._layer_name
        for i in range(len(lst)):
            if lst[i] == self._name:
                if self._stackup.stackup_layers[lst[i - 1]]._layer_type == "dielectric":
                    below_layer = self._stackup.stackup_layers[lst[i - 1]]
                    break
                else:
                    self._app.logger.error("The layer below the selected one must be of dielectric type")
                    return False
        created_patch = Patch(
            self._app,
            frequency,
            patch_width,
            signal_layer=self,
            dielectric_layer=below_layer,
            patch_length=patch_length,
            patch_position_x=patch_position_x,
            patch_position_y=patch_position_y,
            patch_name=patch_name,
            axis=axis,
        )
        self._obj_3d.append(created_patch.aedt_object)
        self._stackup._object_list.append(created_patch)
        created_patch.aedt_object.group_name = "Layer_{}".format(self._name)
        return created_patch

    @pyaedt_function_handler()
    def ml_patch(
        self,
        frequency,
        patch_width,
        patch_position_x=0,
        patch_position_y=0,
        patch_name=None,
        axis="X",
    ):
        """Create a new parametric patch using machine learning algorithm rather than analytic formulas.

        Parameters
        ----------
        frequency : float
            Frequency value for patch calculation in Hz.
        patch_width : float
            Patch width.
        patch_length : float
            Patch Length.
        patch_position_x : float, optional
            Patch start x position.
        patch_position_y : float, optional
            Patch start y position.
        patch_name : str, optional
            Patch name.
        axis : str, optional
            Line orientation axis.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Patch`
        """
        if not patch_name:
            patch_name = generate_unique_name("{}_patch".format(self._name), n=3)
        lst = self._stackup._layer_name
        for i in range(len(lst)):
            if lst[i] == self._name:
                if self._stackup.stackup_layers[lst[i - 1]]._layer_type == "dielectric":
                    below_layer = self._stackup.stackup_layers[lst[i - 1]]
                    break
                else:
                    self._app.logger.error("The layer below the selected one must be of dielectric type")
                    return False
        created_patch = MachineLearningPatch(
            self._app,
            frequency,
            patch_width,
            signal_layer=self,
            dielectric_layer=below_layer,
            patch_position_x=patch_position_x,
            patch_position_y=patch_position_y,
            patch_name=patch_name,
            axis=axis,
        )
        self._obj_3d.append(created_patch.aedt_object)
        self._stackup._object_list.append(created_patch)
        created_patch.aedt_object.group_name = "Layer_{}".format(self._name)
        return created_patch

    @pyaedt_function_handler()
    def add_trace(
        self,
        line_width,
        line_length,
        is_electrical_length=False,
        is_impedance=False,
        line_position_x=0,
        line_position_y=0,
        line_name=None,
        axis="X",
        reference_system=None,
        frequency=1e9,
    ):
        """Create a trace.

        Parameters
        ----------
        line_width : float
            Line width. It can be the physical width or the line impedance.
        line_length : float
            Line length. It can be the physical length or the electrical length.
        is_electrical_length : bool, optional
            Whether the line length is an electrical length or a physical length. The default
            is ``False``, which means it is a physical length.
        is_impedance : bool, optional
            Whether the line width is an impedance. The default is ``False``, in which case
            the line width is a geometrical value.
        line_position_x : float, optional
            Line center start x position. The default is ``0``.
        line_position_y : float, optional
            Line center start y position. The default is ``0``.
        line_name : str, optional
            Line name. The default is ``None``.
        axis : str, optional
            Line orientation axis. The default is ``"X"``.
        reference_system : str, optional
            Line reference system. The default is ``None``, in which case a new coordinate
            system is created.
        frequency : float, optional
            Frequency value for the line calculation in Hz. The default is ``1e9``.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Line`
        """
        if not line_name:
            line_name = generate_unique_name("{0}_line".format(self._name), n=3)
        dielectric_layer = None
        for v in list(self._stackup._stackup.values()):
            if v._index == self._index - 1:
                dielectric_layer = v
                break
        if dielectric_layer is None:
            self._app.logger.error("There is no layer under this layer.")

        created_line = Trace(
            self._app,
            frequency,
            line_width if is_impedance else None,
            line_width if not is_impedance else None,
            self,
            dielectric_layer,
            line_length=line_length if not is_electrical_length else None,
            line_electrical_length=line_length if is_electrical_length else None,
            line_position_x=line_position_x,
            line_position_y=line_position_y,
            line_name=line_name,
            reference_system=reference_system,
            axis=axis,
        )
        created_line.aedt_object.group_name = "Layer_{}".format(self._name)
        self._obj_3d.append(created_line.aedt_object)
        self._stackup._object_list.append(created_line)
        return created_line

    @pyaedt_function_handler()
    def add_polygon(self, points, material="copper", is_void=False, poly_name=None):
        """Create a polygon.

        Parameters
        ----------
        points : list
            Points list of [x,y] coordinates.
        material : str, optional
            Material name. The default is ``"copper"``.
        is_void : bool, optional
            Whether the polygon is a void. The default is ``False``.
            On ground layers, it will act opposite of the Boolean value because the ground
            is negative.
        poly_name : str, optional
            Polygon name. The default is ``None``.

        Returns
        -------

        """
        if not poly_name:
            poly_name = generate_unique_name("{0}_poly".format(self._name), n=3)
        polygon = Polygon(
            self._app,
            points,
            thickness=self._thickness,
            signal_layer_name=self._name,
            mat_name=material,
            is_void=is_void,
            poly_name=poly_name,
        )
        polygon.aedt_object.group_name = "Layer_{}".format(self._name)

        if self._layer_type == "ground":
            if not is_void:
                if polygon.aedt_object.is3d:
                    self._app.modeler[self._name].subtract(polygon.aedt_object, True)
                    polygon.aedt_object.material_name = self.filling_material_name
                else:
                    self._app.modeler[self._name].subtract(polygon.aedt_object, False)
                    return True
        elif is_void:
            if polygon.aedt_object.is3d:
                self._app.modeler.subtract(self._obj_3d, polygon.aedt_object, True)
                polygon.aedt_object.material_name = self.filling_material_name
            else:
                self._app.modeler[self._name].subtract(polygon.aedt_object, False)
                return True
        else:
            self._app.modeler.subtract(self._obj_3d[0], polygon.aedt_object, True)
            self._obj_3d.append(polygon.aedt_object)
            self._stackup._object_list.append(polygon)
        return polygon


class PadstackLayer(object):
    """Provides a data class for the definition of a padstack layer and relative pad and antipad values."""

    def __init__(self, padstack, layer_name, elevation, thickness):
        self._padstack = padstack
        self._layer_name = layer_name
        self._layer_elevation = elevation
        self._layer_thickness = thickness
        self._pad_radius = 1
        self._antipad_radius = 2
        self._units = "mm"


class Padstack(object):
    """Padstack Class member of Stackup3D."""

    def __init__(self, app, stackup, name, material="copper"):
        self._app = app
        self._stackup = stackup
        self.name = name
        self._padstacks_by_layer = OrderedDict({})
        self._vias_objects = []
        self._num_sides = 16
        self._plating_ratio = 1
        v = None
        k = None
        for k, v in self._stackup.stackup_layers.items():
            if not self._padstacks_by_layer and v._layer_type == "dielectric":
                continue
            self._padstacks_by_layer[k] = PadstackLayer(self, k, v.elevation, v.thickness)
        if v and v._layer_type == "dielectric":
            del self._padstacks_by_layer[k]
        self._padstacks_material = material

    @property
    def plating_ratio(self):
        """Plating ratio between 0 and 1.

        Returns
        -------
        float
        """
        return self._plating_ratio

    @plating_ratio.setter
    def plating_ratio(self, val):
        if isinstance(val, (float, int)) and val > 0 and val <= 1:
            self._plating_ratio = val
        elif isinstance(val, str):
            self._plating_ratio = val
        else:
            self._app.logger.error("Plating has to be between 0 and 1")

    @property
    def num_sides(self):
        """Number of sides on the circle, which is 0 for a true circle.

        Returns
        -------
        int
        """
        return self._num_sides

    @num_sides.setter
    def num_sides(self, val):
        self._num_sides = val

    @pyaedt_function_handler()
    def set_all_pad_value(self, value):
        """Set all pads in all layers to a specified value.

        Parameters
        ----------
        value : float
            Pad radius.

        Returns
        -------
        bool
            "True`` when successful, ``False`` when failed.
        """
        for v in list(self._padstacks_by_layer.values()):
            v._pad_radius = value
        return True

    @pyaedt_function_handler()
    def set_all_antipad_value(self, value):
        """Set all antipads in all layers to a specified value.

        Parameters
        ----------
        value : float
            Pad radius.

        Returns
        -------
        bool
             "True`` when successful, ``False`` when failed.
        """
        for v in list(self._padstacks_by_layer.values()):
            v._antipad_radius = value
        return True

    @pyaedt_function_handler()
    def set_start_layer(self, layer):
        """Set the start layer to a specified value.

        Parameters
        ----------
        layer : str
            Layer name.

        Returns
        -------
        bool
             "True`` when successful, ``False`` when failed.

        """
        found = False
        new_stackup = OrderedDict({})
        for k, v in self._stackup.stackup_layers.items():
            if k == layer:
                found = True
            if found and layer not in self._padstacks_by_layer:
                new_stackup[k] = PadstackLayer(self, k, v.elevation)
            elif found:
                new_stackup[k] = self._padstacks_by_layer[k]
        self._padstacks_by_layer = new_stackup
        return True

    @pyaedt_function_handler()
    def set_stop_layer(self, layer):
        """Set the stop layer to a specified value.

        Parameters
        ----------
        layer : str
            Layer name.

        Returns
        -------
        bool
             "True`` when successful, ``False`` when failed.

        """
        found = False
        new_stackup = OrderedDict({})
        for k in list(self._stackup.stackup_layers.keys()):
            if k == layer:
                found = True
            if not found and k in list(self._padstacks_by_layer.keys()):
                new_stackup[k] = self._padstacks_by_layer[k]
        self._padstacks_by_layer = new_stackup

    @pyaedt_function_handler()
    def add_via(self, position_x=0, position_y=0, instance_name=None, reference_system=None):
        """Insert a new via on this padstack.

        Parameters
        ----------
        position_x : float, optional
            Center x position. The default is ``0``.
        position_y : float, optional
            Center y position. The default is ``0``.
        instance_name : str, optional
            Via name. The default is ``None``.
        reference_system : str, optional
            Whether to use an existing reference system or create a new one. The default
            is ``None``, in which case a new reference system is created.

        Returns
        -------
        :class:`pyaedt.modeler.Object3d.Object3d`
            Object created.
        """
        if not instance_name:
            instance_name = generate_unique_name("{}_".format(self.name), n=3)
            if reference_system:
                self._app.modeler.set_working_coordinate_system(reference_system)
                self._reference_system = reference_system
            else:
                self._app.modeler.create_coordinate_system(
                    origin=[0, 0, 0], reference_cs="Global", name=instance_name + "_CS"
                )
                self._app.modeler.set_working_coordinate_system(instance_name + "_CS")
                self._reference_system = instance_name + "_CS"

            first_el = None
            cyls = []
            for v in list(self._padstacks_by_layer.values()):
                if not first_el:
                    first_el = v._layer_elevation
                else:
                    position_x = self._app.modeler._arg_with_dim(position_x)
                    position_y = self._app.modeler._arg_with_dim(position_y)
                    cyls.append(
                        self._app.modeler.create_cylinder(
                            "Z",
                            [position_x, position_y, v._layer_elevation.name],
                            v._pad_radius,
                            v._layer_thickness.name,
                            matname=self._padstacks_material,
                            name=instance_name,
                            numSides=self._num_sides,
                        )
                    )
                    if self.plating_ratio < 1:
                        hole = self._app.modeler.create_cylinder(
                            "Z",
                            [position_x, position_y, v._layer_elevation.name],
                            "{}*{}".format(self._app.modeler._arg_with_dim(v._pad_radius), 1 - self.plating_ratio),
                            v._layer_thickness.name,
                            matname=self._padstacks_material,
                            name=instance_name,
                            numSides=self._num_sides,
                        )
                        cyls[-1].subtract(hole, False)
                    anti = self._app.modeler.create_cylinder(
                        "Z",
                        [position_x, position_y, v._layer_elevation.name],
                        v._antipad_radius,
                        v._layer_thickness.name,
                        matname="air",
                        name=instance_name + "_antipad",
                    )
                    self._app.modeler.subtract(
                        self._stackup._signal_list + self._stackup._ground_list + self._stackup._dielectric_list,
                        anti,
                        False,
                    )
                    first_el = v._layer_elevation
            if len(cyls) > 1:
                self._app.modeler.unite(cyls)
            self._vias_objects.append(cyls[0])
            cyls[0].group_name = "Vias"
            self._stackup._vias.append(self)
            return cyls[0]


class Stackup3D(object):
    """Main Stackup3D Class."""

    def __init__(self, application):
        self._app = application
        self._layer_name = []
        self._layer_position = []
        self._dielectric_list = []
        self._dielectric_name_list = []
        self._ground_list = []
        self._ground_name_list = []
        self._ground_fill_material = []
        self._signal_list = []
        self._signal_name_list = []
        self._signal_material = []
        self._object_list = []
        self._vias = []
        self._end_of_stackup3D = NamedVariable(self._app, "StackUp_End", "0mm")
        self._z_position_offset = 0
        self._first_layer_position = "layer_1_position"
        self._shifted_index = 0
        self._stackup = OrderedDict({})
        self._start_position = NamedVariable(self._app, self._first_layer_position, "0mm")
        self._dielectric_x_position = NamedVariable(self._app, "dielectric_x_position", "0mm")
        self._dielectric_y_position = NamedVariable(self._app, "dielectric_y_position", "0mm")
        self._dielectric_width = NamedVariable(self._app, "dielectric_width", "1000mm")
        self._dielectric_length = NamedVariable(self._app, "dielectric_length", "1000mm")
        self._padstacks = []

    @property
    def padstacks(self):
        """List of padstacks created.

        Returns
        -------
        List
        """
        return self._padstacks

    @property
    def dielectrics(self):
        """List of dielectrics created.

        Returns
        -------
        List
        """
        return self._dielectric_list

    @property
    def grounds(self):
        """List of grounds created.

        Returns
        -------
        List
        """
        return self._ground_list

    @property
    def signals(self):
        """List of signals created.

        Returns
        -------
        List
        """
        return self._signal_list

    @property
    def objects(self):
        """List of obects created.

        Returns
        -------
        List
        """
        return self._object_list

    @property
    def objects_by_layer(self):
        """List of padstacks created.

        Returns
        -------
        List
        """
        objs = {}
        for obj in self.objects:
            if objs.get(obj.layer_name, None):
                objs[obj.layer_name].append(obj)
            else:
                objs[obj.layer_name] = [obj]
        return objs

    @property
    def start_position(self):
        """Variable containing the start position.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
        """
        return self._start_position

    @start_position.setter
    def start_position(self, expression):
        self._start_position.expression = expression

    @property
    def dielectric_x_position(self):
        """Stackup x origin.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._dielectric_x_position

    @dielectric_x_position.setter
    def dielectric_x_position(self, expression):
        self._dielectric_x_position.expression = expression

    @property
    def dielectric_y_position(self):
        """Stackup y origin.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._dielectric_x_position

    @dielectric_y_position.setter
    def dielectric_y_position(self, expression):
        self._dielectric_y_position.expression = expression

    @property
    def dielectric_width(self):
        """Stackup width.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._dielectric_width

    @dielectric_width.setter
    def dielectric_width(self, expression):
        self._dielectric_width.expression = expression

    @property
    def dielectric_length(self):
        """Stackup length.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._dielectric_length

    @dielectric_length.setter
    def dielectric_length(self, expression):
        self._dielectric_length.expression = expression

    @property
    def layer_names(self):
        """List of all layer names.

        Returns
        -------
        list
        """
        return self._layer_name

    @property
    def layer_positions(self):
        """List of all layer positions.

        Returns
        -------
        List
        """
        return self._layer_position

    @property
    def stackup_layers(self):
        """Dictionary of all stackup layers.

        Returns
        -------
        dict
        """
        return self._stackup

    @property
    def z_position_offset(self):
        """Elevation.

        Returns
        -------

        """
        return self._z_position_offset

    @pyaedt_function_handler()
    def add_padstack(self, name, material="copper"):
        """Add a new padstack definition.

        Parameters
        ----------
        name : str
            padstack name
        material : str, optional
            Padstack material. The default is ``"copper"``.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Padstack`
        """
        p = Padstack(self._app, self, name, material)
        self._padstacks.append(p)
        return p

    @pyaedt_function_handler()
    def add_layer(self, name, layer_type="S", material="copper", thickness=0.035, fill_material="FR4_epoxy"):
        """Add a new layer to the stackup.
        The new layer can be a signal (S), ground (G), or dielectric (D).
        The layer is entirely filled with the specified fill material. Anything will be drawn
        wmaterial.

        Parameters
        ----------
        name : str
            Layer name.
        layer_type : str, optional
            Layer type. Options are ``"S"``, ``"D"``, and ``"G"``. The default is ``"S"``.
        material : str, optional
            Material name. The default is ``"copper"``. The material will be parametrized.
        thickness : float, optional
            Thickness value. The default is ``0.035``. The thickness will be parametrized.
        fill_material : str, optional
            Fill material name. The default is ``"FR4_epoxy"``. The fill material will be
            parametrized. This parameter is not valid for dielectrics.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Layer3D`
            Layer object.
        """
        self._shifted_index += 1
        if not layer_type:
            raise ValueError("Layer type has to be an S, D, or G string.")
        self._layer_name.append(name)

        lay = Layer3D(
            stackup=self,
            app=self._app,
            name=name,
            layer_type=layer_type,
            material=material,
            thickness=thickness,
            fill_material=fill_material,
            index=self._shifted_index,
        )
        self._layer_position_manager(lay)
        if layer_type == "D":
            self._dielectric_list.extend(lay._obj_3d)
            self._dielectric_name_list.append(lay._name)
            lay._obj_3d[-1].transparency = "0.8"
        elif layer_type == "G":
            self._ground_list.extend(lay._obj_3d)
            self._ground_name_list.append(lay._name)
            self._ground_fill_material.append(lay._fill_material)
            lay._obj_3d[-1].transparency = "0.6"
            lay._obj_3d[-1].color = (255, 0, 0)

        elif layer_type == "S":
            self._signal_list.extend(lay._obj_3d)
            self._signal_name_list.append(lay._name)
            self._signal_material.append(lay._material_name)
            # With the function _layer_position_manager i think this part is not needed anymore or has to be reworked
            lay._obj_3d[-1].transparency = "0.8"
        self._stackup[lay._name] = lay
        return lay

    @pyaedt_function_handler()
    def add_signal_layer(self, name, material="copper", thickness=0.035, fill_material="FR4_epoxy"):
        """Add a new ground layer to the stackup.
        A signal layer is positive. The layer is entirely filled with the fill material.
        Anything will be drawn wmaterial.

        Parameters
        ----------
        name : str
            Layer name.
        material : str
            Material name. Material will be parametrized.
        thickness : float
            Thickness value. Thickness will be parametrized.
        fill_material : str
            Fill Material name. Material will be parametrized.=
        material : str, optional
            Material name. Material will be parametrized. Default value is `"copper"`.
        thickness : float, optional
            Thickness value. Thickness will be parametrized. Default value is `0.035`.
        fill_material : str, optional
            Fill material name. Material will be parametrized. Default value is `"FR4_epoxy"`.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Layer3D`
            Layer object.
        """
        return self.add_layer(
            name=name, layer_type="S", material=material, thickness=thickness, fill_material=fill_material
        )

    @pyaedt_function_handler()
    def add_dielectric_layer(
        self,
        name,
        material="FR4_epoxy",
        thickness=0.035,
    ):
        """Add a new dielectric layer to the stackup.

        Parameters
        ----------
        name : str
            Layer name.
        material : str
            Material name. The default is ``"FR4_epoxy"``. The material will be parametrized.
        thickness : float, optional
            Thickness value. The default is ``0.035``. The thickness will be parametrized.


        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Layer3D`
            Layer 0bject.
        """
        return self.add_layer(name=name, layer_type="D", material=material, thickness=thickness, fill_material=None)

    @pyaedt_function_handler()
    def add_ground_layer(self, name, material="copper", thickness=0.035, fill_material="air"):
        """Add a new ground layer to the stackup. A ground layer is negative.
        The layer is entirely filled with  metal. Any polygon will draw a void in it.

        Parameters
        ----------
        name : str
            Layer name.
        material : str
            Material name. Material will be parametrized.
        thickness : float
            Thickness value. Thickness will be parametrized.
        fill_material : str
            Fill Material name. Material will be parametrized.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Layer3D`
            Layer Object.
        """
        return self.add_layer(
            name=name, layer_type="G", material=material, thickness=thickness, fill_material=fill_material
        )

    @pyaedt_function_handler()
    def _layer_position_manager(self, layer):
        """

        Parameters
        ----------
        layer

        Returns
        -------

        """
        previous_layer_end = self._end_of_stackup3D.expression

        layer.elevation.expression = previous_layer_end
        if layer.thickness:
            self._end_of_stackup3D.expression = layer.elevation.name + " + " + layer.thickness.name
        else:
            self._end_of_stackup3D.expression = layer.elevation.name

    # if we call this function instantiation of the Layer, the first call, previous_layer_end is "0mm", and
    # layer.position.expression is also "0mm" and self._end_of_stackup becomes the first layer.position + thickness
    # if it has thickness, and so the second call, previous_layer_end is the previous layer position + thickness
    # so the current layer position is the previous_layer_end and the end_of_stackup is the current layer position +
    # thickness, and we just need to call this function after the construction of a layer3D.

    @pyaedt_function_handler()
    def resize(self, percentage_offset):
        """Resize the stackup around objects created by a percentage offset.

        Parameters
        ----------
        percentage_offset : float
            Offset of resize. The value must be greater than 0.

        Returns
        -------
        bool
        """
        list_of_2d_points = []
        list_of_x_coordinates = []
        list_of_y_coordinates = []
        for obj3d in self._object_list:
            points_list_by_object = obj3d.points_on_layer
            list_of_2d_points = points_list_by_object + list_of_2d_points
        for via in self._vias:
            for v in via._vias_objects:
                list_of_x_coordinates.append(v.bounding_box[0] - v.bounding_dimension[0])
                list_of_x_coordinates.append(v.bounding_box[3] - v.bounding_dimension[0])
                list_of_y_coordinates.append(v.bounding_box[1] - v.bounding_dimension[1])
                list_of_y_coordinates.append(v.bounding_box[4] - v.bounding_dimension[1])
                list_of_x_coordinates.append(v.bounding_box[0] + v.bounding_dimension[0])
                list_of_x_coordinates.append(v.bounding_box[4] + v.bounding_dimension[0])
                list_of_y_coordinates.append(v.bounding_box[4] + v.bounding_dimension[1])
                list_of_y_coordinates.append(v.bounding_box[1] + v.bounding_dimension[1])
        for point in list_of_2d_points:
            list_of_x_coordinates.append(point[0])
            list_of_y_coordinates.append(point[1])
        maximum_x = max(list_of_x_coordinates)
        minimum_x = min(list_of_x_coordinates)
        maximum_y = max(list_of_y_coordinates)
        minimum_y = min(list_of_y_coordinates)
        variation_x = abs(maximum_x - minimum_x)
        variation_y = abs(maximum_y - minimum_y)
        self._app["dielectric_x_position"] = str(minimum_x - variation_x * percentage_offset / 100) + "mm"
        self._app["dielectric_y_position"] = str(minimum_y - variation_y * percentage_offset / 100) + "mm"
        self._app["dielectric_length"] = str(maximum_x - minimum_x + 2 * variation_x * percentage_offset / 100) + "mm"
        self._app["dielectric_width"] = str(maximum_y - minimum_y + 2 * variation_y * percentage_offset / 100) + "mm"
        return True

    def resize_around_element(self, element, percentage_offset=0.25):
        """Resize the stackup around objects and make it parametrize.

        Parameters
        ----------
        element : :class:`pyaedt.modeler.stackup_3d.Patch
            Element around which the resizing is done.
        percentage_offset : float, optional
            Offset of resize. Value accepted are greater than 0. O.25 by default.

        Returns
        -------
        bool
        """
        self._app["dielectric_x_position"] = (
            element.position_x.name + " - " + element.length.name + " * " + str(percentage_offset)
        )
        self._app["dielectric_y_position"] = (
            element.position_y.name + " - " + element.width.name + " * (0.5 + " + str(percentage_offset) + ")"
        )
        self._app["dielectric_length"] = element.length.name + " * (1 + " + str(percentage_offset) + " * 2)"
        self._app["dielectric_width"] = element.width.name + " * (1 + " + str(percentage_offset) + " * 2)"
        return True


class CommonObject(object):
    """CommonObject Class in Stackup3D."""

    def __init__(self, application):
        self._application = application
        self._name = None
        self._dielectric_layer = None
        self._signal_layer = None
        self._aedt_object = None
        self._layer_name = None
        self._layer_number = None
        self._material_name = None
        self._reference_system = None

    @property
    def reference_system(self):
        """Coordinate system of the object.

        Returns
        -------
        str
        """
        return self._reference_system

    @property
    def dielectric_layer(self):
        """Dielectric layer that the object belongs to.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Layer3D`
        """
        return self._dielectric_layer

    @property
    def signal_layer(self):
        """Signal layer that the object belongs to.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.Layer3D`
        """
        return self._signal_layer

    @property
    def name(self):
        """Object name.

        Returns
        -------
        str
        """
        return self._name

    @property
    def application(self):
        """App object."""
        return self._application

    @property
    def aedt_object(self):
        """PyAEDT object 3D.

        Returns
        -------
        :class:`pyaedt.modeler.Object3d.Object3d`
        """
        return self._aedt_object

    @property
    def layer_name(self):
        """Layer name.

        Returns
        -------
        str
        """
        return self._layer_name

    @property
    def layer_number(self):
        """Layer ID.

        Returns
        -------
        int
        """
        return self._layer_number

    @property
    def material_name(self):
        """Material name.

        Returns
        -------
        str
        """
        return self._material_name

    @property
    def points_on_layer(self):
        """Object bounding box.

        Returns
        -------
        List
            List of [x,y] coordinate of the bounding box.
        """
        bb = self._aedt_object.bounding_box
        return [[bb[0], bb[1]], [bb[0], bb[4]], [bb[3], bb[4]], [bb[3], bb[1]]]


class Patch(CommonObject, object):
    """Patch Class in Stackup3D."""

    def __init__(
        self,
        application,
        frequency,
        patch_width,
        signal_layer,
        dielectric_layer,
        patch_length=None,
        patch_position_x=0,
        patch_position_y=0,
        patch_name="patch",
        reference_system=None,
        axis="X",
    ):
        CommonObject.__init__(self, application)
        self._frequency = NamedVariable(application, patch_name + "_frequency", str(frequency) + "Hz")
        self._signal_layer = signal_layer
        self._dielectric_layer = dielectric_layer
        self._substrate_thickness = dielectric_layer.thickness
        self._width = NamedVariable(application, patch_name + "_width", application.modeler._arg_with_dim(patch_width))
        self._position_x = NamedVariable(
            application, patch_name + "_position_x", application.modeler._arg_with_dim(patch_position_x)
        )
        self._position_y = NamedVariable(
            application, patch_name + "_position_y", application.modeler._arg_with_dim(patch_position_y)
        )
        self._position_z = signal_layer.elevation
        self._dielectric_layer = dielectric_layer
        self._signal_layer = signal_layer
        self._dielectric_material = dielectric_layer.material
        self._material_name = signal_layer.material_name
        self._layer_name = signal_layer.name
        self._layer_number = signal_layer.number
        self._name = patch_name
        self._patch_thickness = signal_layer.thickness
        self._application = application
        self._aedt_object = None
        try:
            self._permittivity = NamedVariable(
                application, patch_name + "_permittivity", float(self._dielectric_material.permittivity.value)
            )
        except ValueError:
            self._permittivity = NamedVariable(
                application,
                patch_name + "_permittivity",
                float(application.variable_manager[self._dielectric_material.permittivity.value].value),
            )
        if isinstance(patch_length, float) or isinstance(patch_length, int):
            self._length = NamedVariable(
                application, patch_name + "_length", application.modeler._arg_with_dim(patch_length)
            )
            self._effective_permittivity = self._effective_permittivity_calcul
            self._wave_length = self._wave_length_calcul
        elif patch_length is None:
            self._effective_permittivity = self._effective_permittivity_calcul
            self._added_length = self._added_length_calcul
            self._wave_length = self._wave_length_calcul
            self._length = self._length_calcul
        self._impedance_l_w, self._impedance_w_l = self._impedance_calcul
        if reference_system:
            application.modeler.set_working_coordinate_system(reference_system)
            if axis == "X":
                start_point = [
                    "{0}_position_x".format(self._name),
                    "{0}_position_y-{0}_width/2".format(self._name),
                    0,
                ]
            else:
                start_point = [
                    "{0}_position_x-{0}_width/2".format(self._name),
                    "{}_position_y".format(self._name),
                    0,
                ]
            self._reference_system = reference_system
        else:
            application.modeler.create_coordinate_system(
                origin=[
                    "{0}_position_x".format(patch_name),
                    "{}_position_y".format(patch_name),
                    signal_layer.elevation.name,
                ],
                reference_cs="Global",
                name=patch_name + "_CS",
            )
            if axis == "X":
                start_point = [0, "-{}_width/2".format(patch_name), 0]

            else:
                start_point = ["-{}_width/2".format(patch_name), 0, 0]
            application.modeler.set_working_coordinate_system(patch_name + "_CS")

            self._reference_system = patch_name + "_CS"
        if signal_layer.thickness:
            self._aedt_object = application.modeler.create_box(
                position=start_point,
                dimensions_list=[
                    "{}_length".format(patch_name),
                    "{}_width".format(patch_name),
                    signal_layer.thickness.name,
                ],
                name=patch_name,
                matname=signal_layer.material_name,
            )
        else:
            self._aedt_object = application.modeler.create_rectangle(
                position=start_point,
                dimension_list=[self.length.name, self.width.name],
                name=patch_name,
                matname=signal_layer.material_name,
            )
            application.assign_coating(self._aedt_object.name, signal_layer.material)
        application.modeler.set_working_coordinate_system("Global")
        application.modeler.subtract(blank_list=[signal_layer.name], tool_list=[patch_name], keepOriginals=True)

    @property
    def frequency(self):
        """Model frequency.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._frequency

    @property
    def substrate_thickness(self):
        """Substrate thickness.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._substrate_thickness

    @property
    def width(self):
        """Width.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._width

    @property
    def position_x(self):
        """Starting position X.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._position_x

    @property
    def position_y(self):
        """Starting position Y.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._position_y

    @property
    def permittivity(self):
        """Permittivity.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._permittivity

    @property
    def _permittivity_calcul(self):
        """Permittivity calculation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        self._permittivity = self.application.materials[self._dielectric_material].permittivity
        return self._permittivity

    @property
    def effective_permittivity(self):
        """Effective permittivity.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._effective_permittivity

    @property
    def _effective_permittivity_calcul(self):
        # "(substrat_permittivity + 1)/2 + (substrat_permittivity -
        # 1)/(2 * sqrt(1 + 10 * substrate_thickness/patch_width))"
        er = self._permittivity.name
        h = self._substrate_thickness.name
        w = self._width.name
        patch_eff_permittivity_formula = "(" + er + "+ 1)/2 + (" + er + "- 1)/(2 * sqrt(1 + 10 * " + h + "/" + w + "))"
        self._effective_permittivity = NamedVariable(
            self.application, self._name + "_eff_permittivity", patch_eff_permittivity_formula
        )
        return self._effective_permittivity

    @property
    def added_length(self):
        """Added length calculation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        return self._added_length

    @property
    def _added_length_calcul(self):
        """Added length calculation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable object.
        """
        # "0.412 * substrate_thickness * (patch_eff_permittivity + 0.3) * (patch_width/substrate_thickness + 0.264)"
        # " / ((patch_eff_permittivity - 0.258) * (patch_width/substrate_thickness + 0.813)) "

        er_e = self._effective_permittivity.name
        h = self._substrate_thickness.name
        w = self._width.name
        patch_added_length_formula = (
            "0.412 * " + h + " * (" + er_e + " + 0.3) * (" + w + "/" + h + " + 0.264)/"
            "((" + er_e + " - 0.258) * (" + w + "/" + h + " + 0.813))"
        )
        self._added_length = NamedVariable(self.application, self._name + "_added_length", patch_added_length_formula)
        return self._added_length

    @property
    def wave_length(self):
        """Wave length.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._wave_length

    @property
    def _wave_length_calcul(self):
        """Wave Length Calutation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        # "c0 * 1000/(patch_frequency * sqrt(patch_eff_permittivity))"
        f = self._frequency.name
        er_e = self._effective_permittivity.name
        patch_wave_length_formula = "(c0 * 1000/(" + f + "* sqrt(" + er_e + ")))mm"
        self._wave_length = NamedVariable(
            self.application,
            self._name + "_wave_length",
            self.application.modeler._arg_with_dim(patch_wave_length_formula),
        )
        return self._wave_length

    @property
    def length(self):
        """Length.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._length

    @property
    def _length_calcul(self):
        """Length Calutation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        # "patch_wave_length / 2 - 2 * patch_added_length"
        d_l = self._added_length.name
        lbd = self._wave_length.name
        patch_length_formula = lbd + "/2" + " - 2 * " + d_l
        self._length = NamedVariable(self.application, self._name + "_length", patch_length_formula)
        return self._length

    @property
    def impedance(self):
        """Impedance.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._impedance_l_w, self._impedance_w_l

    @property
    def _impedance_calcul(self):
        """Impedance Calculation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        # "45 * (patch_wave_length/patch_width * sqrt(patch_eff_permittivity)) ** 2"
        # "60 * patch_wave_length/patch_width * sqrt(patch_eff_permittivity)"
        er_e = self._effective_permittivity.name
        lbd = self._wave_length.name
        w = self._width.name
        patch_impedance_formula_l_w = "45 * (" + lbd + "/" + w + "* sqrt(" + er_e + ")) ** 2"
        patch_impedance_formula_w_l = "60 * " + lbd + "/" + w + "* sqrt(" + er_e + ")"
        self._impedance_l_w = NamedVariable(
            self.application, self._name + "_impedance_l_w", patch_impedance_formula_l_w
        )
        self._impedance_w_l = NamedVariable(
            self.application, self._name + "_impedance_w_l", patch_impedance_formula_w_l
        )
        self.application.logger.warning(
            "The closer the ratio between wave length and the width is to 1,"
            " the less correct the impedance calculation is"
        )
        return self._impedance_l_w, self._impedance_w_l

    def create_lumped_port(self, reference_layer, opposite_side=False, port_name=None, axisdir=None):
        """Create a parametrized lumped port.

        Parameters
        ----------

        reference_layer : class:`pyaedt.modeler.stackup_3d.Layer3D
            The reference layer, in most cases the ground layer.
        opposite_side : bool, optional
            Change the side where the port is created.
        port_name : str, optional
            Name of the lumped port.
        axisdir : int or :class:`pyaedt.application.Analysis.Analysis.AxisDir`, optional
            Position of the port. It should be one of the values for ``Application.AxisDir``,
            which are: ``XNeg``, ``YNeg``, ``ZNeg``, ``XPos``, ``YPos``, and ``ZPos``.
            The default is ``Application.AxisDir.XNeg``.
        Returns
        -------
        bool
        """
        string_position_x = self.position_x.name
        if opposite_side:
            string_position_x = self.position_x.name + " + " + self.length.name
        string_position_y = self.position_y.name + " - " + self.width.name + "/2"
        string_position_z = reference_layer.elevation.name
        string_width = self.width.name
        string_length = (
            self._signal_layer.elevation.name
            + " + "
            + self._signal_layer.thickness.name
            + " - "
            + reference_layer.elevation.name
        )
        port = self.application.modeler.create_rectangle(
            csPlane=constants.PLANE.YZ,
            position=[string_position_x, string_position_y, string_position_z],
            dimension_list=[string_width, string_length],
            name=self.name + "_port",
            matname=None,
        )
        if self.application.solution_type == "Modal":
            if axisdir is None:
                axisdir = self.application.AxisDir.ZPos
            port = self.application.create_lumped_port_to_sheet(port.name, portname=port_name, axisdir=axisdir)
        elif self.application.solution_type == "Terminal":
            port = self.application.create_lumped_port_to_sheet(
                port.name, portname=port_name, reference_object_list=[reference_layer.name]
            )
        return port


class Trace(CommonObject, object):
    """Provides a class to create a trace in stackup."""

    def __init__(
        self,
        application,
        frequency,
        line_impedance,
        line_width,
        signal_layer,
        dielectric_layer,
        line_length=None,
        line_electrical_length=90,
        line_position_x=0,
        line_position_y=0,
        line_name="line",
        reference_system=None,
        axis="X",
    ):
        CommonObject.__init__(self, application)
        self._frequency = NamedVariable(application, line_name + "_frequency", str(frequency) + "Hz")
        self._signal_layer = signal_layer
        self._dielectric_layer = dielectric_layer
        self._substrate_thickness = dielectric_layer.thickness
        self._position_x = NamedVariable(
            application, line_name + "_position_x", application.modeler._arg_with_dim(line_position_x)
        )
        self._position_y = NamedVariable(
            application, line_name + "_position_y", application.modeler._arg_with_dim(line_position_y)
        )
        self._position_z = signal_layer.elevation
        self._dielectric_material = dielectric_layer.material
        self._material_name = signal_layer.material_name
        self._layer_name = signal_layer.name
        self._layer_number = signal_layer.number
        self._name = line_name
        self._line_thickness = signal_layer.thickness
        self._width = None
        self._width_h_w = None
        self._axis = axis
        try:
            self._permittivity = NamedVariable(
                application, line_name + "_permittivity", float(self._dielectric_material.permittivity.value)
            )
        except ValueError:
            self._permittivity = NamedVariable(
                application,
                line_name + "_permittivity",
                float(application.variable_manager[self._dielectric_material.permittivity.value].value),
            )
        if isinstance(line_width, float) or isinstance(line_width, int):
            self._width = NamedVariable(
                application, line_name + "_width", application.modeler._arg_with_dim(line_width)
            )
            self._effective_permittivity = self._effective_permittivity_calcul
            self._wave_length = self._wave_length_calcul
            self._added_length = self._added_length_calcul
            if isinstance(line_electrical_length, float) or isinstance(line_electrical_length, int):
                self._electrical_length = NamedVariable(
                    application, line_name + "_elec_length", str(line_electrical_length)
                )
                self._length = self._length_calcul
            elif isinstance(line_length, float) or isinstance(line_length, int):
                self._length = NamedVariable(
                    application, line_name + "_length", application.modeler._arg_with_dim(line_length)
                )
                self._electrical_length = self._electrical_length_calcul
            else:
                application.logger.error("line_length must be a float.")
            self._charac_impedance_w_h, self._charac_impedance_h_w = self._charac_impedance_calcul
        elif line_width is None:
            self._charac_impedance = NamedVariable(
                self.application, line_name + "_charac_impedance_h_w", str(line_impedance)
            )
            self._width, self._width_h_w = self._width_calcul
            self._effective_permittivity = self._effective_permittivity_calcul
            self._wave_length = self._wave_length_calcul
            self._added_length = self._added_length_calcul
            if isinstance(line_electrical_length, float) or isinstance(line_electrical_length, int):
                self._electrical_length = NamedVariable(
                    application, line_name + "_elec_length", str(line_electrical_length)
                )
                self._length = self._length_calcul
            elif isinstance(line_length, float) or isinstance(line_length, int):
                self._length = NamedVariable(
                    application, line_name + "_length", application.modeler._arg_with_dim(line_length)
                )
                self._electrical_length = self._electrical_length_calcul
            else:
                application.logger.error("line_length must be a float.")
        if reference_system:
            application.modeler.set_working_coordinate_system(reference_system)
            if axis == "X":
                start_point = [
                    "{0}_position_x".format(self._name),
                    "{0}_position_y-{0}_width/2".format(self._name),
                    0,
                ]
            else:

                start_point = [
                    "{0}_position_x-{0}_width/2".format(self._name),
                    "{}_position_y".format(self._name),
                    0,
                ]
            self._reference_system = reference_system
        else:
            application.modeler.create_coordinate_system(
                origin=[
                    "{}_position_x".format(self._name),
                    "{}_position_y".format(self._name),
                    signal_layer.elevation.name,
                ],
                reference_cs="Global",
                name=line_name + "_CS",
            )
            application.modeler.set_working_coordinate_system(line_name + "_CS")
            if axis == "X":
                start_point = [0, "-{0}_width/2".format(self._name), 0]
            else:
                start_point = ["-{0}_width/2".format(self._name), 0, 0]
            self._reference_system = line_name + "_CS"
        if signal_layer.thickness:
            self._aedt_object = application.modeler.create_box(
                position=start_point,
                dimensions_list=[
                    "{}_length".format(self._name),
                    "{}_width".format(self._name),
                    signal_layer.thickness.name,
                ],
                name=line_name,
                matname=signal_layer.material_name,
            )
        else:
            self._aedt_object = application.modeler.create_rectangle(
                position=start_point,
                dimension_list=["{}_length".format(self._name), "{}_width".format(self._name)],
                name=line_name,
                matname=signal_layer.material_name,
            )
        application.modeler.set_working_coordinate_system("Global")
        application.modeler.subtract(blank_list=[signal_layer.name], tool_list=[line_name], keepOriginals=True)

    @property
    def frequency(self):
        """Frequency.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._frequency

    @property
    def substrate_thickness(self):
        """Substrate Thickness.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._substrate_thickness

    @property
    def width(self):
        """Width.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._width

    @property
    def width_h_w(self):
        """Width H W.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        if self._width_h_w is not None:
            return self._width_h_w

    @property
    def _width_calcul(self):
        """Width calculation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        # if w/h < 2 :
        # a = z * sqrt((er + 1) / 2) / 60 + (0.23 + 0.11 / er) * (er - 1) / (er + 1)
        # w/h = 8 * exp(a) / (exp(2 * a) - 2)
        # else w/h > 2 :
        # b = 377 * pi / (2 * z * sqrt(er))
        # w/h = 2 * (b - 1 - log(2 * b - 1) * (er - 1) * (log(b - 1) + 0.39 - 0.61 / er) / (2 * er)) / pi
        h = self._substrate_thickness.name
        z = self._charac_impedance.name
        er = self._permittivity.name
        a_formula = (
            "("
            + z
            + " * sqrt(("
            + er
            + " + 1)/2)/60 + (0.23 + 0.11/"
            + er
            + ")"
            + " * ("
            + er
            + "- 1)/("
            + er
            + "+ 1))"
        )
        w_div_by_h_inf_2 = "(8 * exp(" + a_formula + ")/(exp(2 * " + a_formula + ") - 2))"

        b_formula = "(377 * pi/(2 * " + z + " * " + "sqrt(" + er + ")))"
        w_div_by_h_sup_2 = (
            "(2 * ("
            + b_formula
            + " - 1 - log(2 * "
            + b_formula
            + " - 1) * ("
            + er
            + " - 1) * (log("
            + b_formula
            + " - 1) + 0.39 - 0.61/"
            + er
            + ")/(2 * "
            + er
            + "))/pi)"
        )

        w_formula_inf = w_div_by_h_inf_2 + " * " + h
        w_formula_sup = w_div_by_h_sup_2 + " * " + h

        self._width_h_w = NamedVariable(self.application, self._name + "_width_h_w", w_formula_inf)
        self._width = NamedVariable(self.application, self._name + "_width", w_formula_sup)

        return self._width, self._width_h_w

    @property
    def position_x(self):
        """Starting Position X.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._position_x

    @property
    def position_y(self):
        """Starting Position Y.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._position_y

    @property
    def permittivity(self):
        """Permittivity.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._permittivity

    @property
    def _permittivity_calcul(self):
        """Permittivity Calutation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        self._permittivity = self.application.materials[self._dielectric_material].permittivity
        return self._permittivity

    @property
    def added_length(self):
        """Added Length Calutation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._added_length

    @property
    def _added_length_calcul(self):
        """Added Length Calutation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        # "0.412 * substrate_thickness * (patch_eff_permittivity + 0.3) * (patch_width/substrate_thickness + 0.264)"
        # " / ((patch_eff_permittivity - 0.258) * (patch_width/substrate_thickness + 0.813)) "

        er_e = self._effective_permittivity.name
        h = self._substrate_thickness.name
        w = self._width.name
        patch_added_length_formula = (
            "0.412 * " + h + " * (" + er_e + " + 0.3) * (" + w + "/" + h + " + 0.264)/"
            "((" + er_e + " - 0.258) * (" + w + "/" + h + " + 0.813))"
        )
        self._added_length = NamedVariable(self.application, self._name + "_added_length", patch_added_length_formula)
        return self._added_length

    @property
    def length(self):
        """Length.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._length

    @property
    def _length_calcul(self):
        """Length Calutation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        # "patch_wave_length / 2 - 2 * patch_added_length"
        d_l = self._added_length.name
        lbd = self._wave_length.name
        e_l = self._electrical_length.name
        line_length_formula = lbd + "* (" + e_l + "/360)" + " - 2 * " + d_l
        self._length = NamedVariable(self.application, self._name + "_length", line_length_formula)
        return self._length

    @property
    def charac_impedance(self):
        """Characteristic Impedance.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._charac_impedance

    @property
    def _charac_impedance_calcul(self):
        """Characteristic Impedance Calutation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        # if w / h > 1: 60 * log(8 * h / w + w / (4 * h)) / sqrt(er_e)
        # if w / h < 1: 120 * pi / (sqrt(er_e) * (w / h + 1.393 + 0.667 * log(w / h + 1.444)))
        w = self._width.name
        h = self._dielectric_layer.thickness.name
        er_e = self.effective_permittivity.name
        charac_impedance_formula_w_h = (
            "60 * log(8 * " + h + "/" + w + " + " + w + "/(4 * " + h + "))/sqrt(" + er_e + ")"
        )
        charac_impedance_formula_h_w = (
            "120 * pi / (sqrt(" + er_e + ") * (" + w + "/" + h + "+ 1.393 + 0.667 * log(" + w + "/" + h + " + 1.444)))"
        )
        self._charac_impedance_w_h = NamedVariable(
            self.application, self._name + "_charac_impedance_w_h", charac_impedance_formula_w_h
        )
        self._charac_impedance_h_w = NamedVariable(
            self.application, self._name + "_charac_impedance_h_w", charac_impedance_formula_h_w
        )
        return self._charac_impedance_w_h, self._charac_impedance_h_w

    @property
    def effective_permittivity(self):
        """Effective Permittivity.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._effective_permittivity

    @property
    def _effective_permittivity_calcul(self):
        """Effective Permittivity Calutation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        # "(substrat_permittivity + 1)/2 +
        # (substrat_permittivity - 1)/(2 * sqrt(1 + 10 * substrate_thickness/patch_width))"
        er = self._permittivity.name
        h = self._substrate_thickness.name
        w = self._width.name
        patch_eff_permittivity_formula = (
            "(" + er + " + 1)/2 + (" + er + " - 1)/(2 * sqrt(1 + 10 * " + h + "/" + w + "))"
        )
        self._effective_permittivity = NamedVariable(
            self.application, self._name + "_eff_permittivity", patch_eff_permittivity_formula
        )
        return self._effective_permittivity

    @property
    def wave_length(self):
        """Wave Length.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._wave_length

    @property
    def _wave_length_calcul(self):
        """Wave Length Calutation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        # "c0 * 1000/(patch_frequency * sqrt(patch_eff_permittivity))"
        # TODO it is currently only available for mm
        f = self._frequency.name
        er_e = self._effective_permittivity.name
        patch_wave_length_formula = "(c0 * 1000/(" + f + "* sqrt(" + er_e + ")))mm"
        self._wave_length = NamedVariable(
            self.application,
            self._name + "_wave_length",
            self.application.modeler._arg_with_dim(patch_wave_length_formula),
        )
        return self._wave_length

    @property
    def electrical_length(self):
        """Electrical Length.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        return self._electrical_length

    @property
    def _electrical_length_calcul(self):
        """Electrical Length calculation.

        Returns
        -------
        :class:`pyaedt.modeler.stackup_3d.NamedVariable`
            Variable Object.
        """
        lbd = self._wave_length.name
        length = self._length.name
        d_l = self._added_length.name
        elec_length_formula = "360 * (" + length + " + 2 * " + d_l + ")/" + lbd
        self._electrical_length = NamedVariable(self.application, self._name + "_elec_length", elec_length_formula)
        return self._electrical_length

    @pyaedt_function_handler()
    def create_lumped_port(self, reference_layer_name, change_side=False):
        """Create a lumped port on the specified line.

        Parameters
        ----------
        reference_layer_name : str
            Name of the layer on which attach the reference.
        change_side : bool, optional
            Either if apply the port on one direction or the opposite. Default it is on Positive side.

        Returns
        -------
        :class:`pyaedt.modules.Boundary.BoundaryObject`
            Boundary object.
        """
        if self._axis == "X":
            if change_side:
                axisdir = self.application.AxisDir.XNeg
            else:
                axisdir = self.application.AxisDir.XPos
        else:
            if change_side:
                axisdir = self.application.AxisDir.YNeg
            else:
                axisdir = self.application.AxisDir.YPos
        p1 = self.application.create_lumped_port_between_objects(
            reference_layer_name, self.aedt_object.name, axisdir=axisdir
        )
        z_elev = ""
        start_count = False
        for k, v in self._signal_layer._stackup.stackup_layers.items():
            if k == reference_layer_name or k == self._signal_layer.name:
                if not start_count:
                    start_count = True
                else:
                    start_count = False
            elif start_count:
                z_elev += "-" + v.thickness.name
        self.application.modeler.oeditor.ChangeProperty(
            [
                "NAME:AllTabs",
                [
                    "NAME:Geometry3DCmdTab",
                    ["NAME:PropServers", self._name + ":Move:1"],
                    ["NAME:ChangedProps", ["NAME:Move Vector", "X:=", "0mm", "Y:=", "0mm", "Z:=", z_elev]],
                ],
            ]
        )
        return p1


class Polygon(CommonObject, object):
    """Polygon Class in Stackup3D."""

    def __init__(
        self,
        application,
        point_list,
        thickness,
        signal_layer_name,
        poly_name="poly",
        mat_name="copper",
        is_void=False,
        reference_system=None,
    ):
        CommonObject.__init__(self, application)

        self._is_void = is_void
        self._layer_name = signal_layer_name
        self._app = application
        pts = []
        for el in point_list:
            pts.append(
                [
                    application.modeler._arg_with_dim(el[0]),
                    application.modeler._arg_with_dim(el[1]),
                    "layer_" + str(signal_layer_name) + "_position",
                ]
            )
        if reference_system:
            application.modeler.set_working_coordinate_system(reference_system)
            self._reference_system = reference_system
        else:
            application.modeler.create_coordinate_system(
                origin=[0, 0, 0], reference_cs="Global", name=poly_name + "_CS"
            )
            application.modeler.set_working_coordinate_system(poly_name + "_CS")

            self._reference_system = poly_name + "_CS"
        self._aedt_object = application.modeler.create_polyline(
            position_list=pts, name=poly_name, matname=mat_name, cover_surface=True
        )
        if thickness:
            if isinstance(thickness, (float, int)):
                application.modeler.sweep_along_vector(self._aedt_object, [0, 0, thickness], draft_type="Natural")
            else:
                application.modeler.sweep_along_vector(self._aedt_object, [0, 0, thickness.name], draft_type="Natural")
        application.modeler.set_working_coordinate_system("Global")

    @property
    def points_on_layer(self):
        """Object Bounding Box.

        Returns
        -------
        List
            List of [x,y] coordinate of bounding box.
        """
        bb = self._aedt_object.bounding_box
        return [[bb[0], bb[1]], [bb[0], bb[4]], [bb[3], bb[4]], [bb[3], bb[1]]]


class MachineLearningPatch(Patch, object):
    """MachineLearningPatch Class in Stackup3D."""

    def __init__(
        self,
        application,
        frequency,
        patch_width,
        signal_layer,
        dielectric_layer,
        patch_position_x=0,
        patch_position_y=0,
        patch_name="patch",
        reference_system=None,
        axis="X",
    ):
        Patch.__init__(
            self,
            application,
            frequency,
            patch_width,
            signal_layer,
            dielectric_layer,
            patch_length=None,
            patch_position_x=patch_position_x,
            patch_position_y=patch_position_y,
            patch_name=patch_name,
            reference_system=reference_system,
            axis=axis,
        )
        if not is_ironpython:
            try:
                joblib
            except NameError:  # pragma: no cover
                raise ImportError("joblib package is needed to run ML.")
            path_file = os.path.dirname(__file__)
            path_folder = os.path.split(path_file)[0]
            training_file = os.path.join(path_folder, "misc", "patch_svr_model_100MHz_1GHz.joblib")
            model = joblib.load(training_file)
            list_for_array = [
                [
                    self.frequency.numeric_value,
                    self.width.numeric_value,
                    self._permittivity.numeric_value,
                    self.dielectric_layer.thickness.numeric_value,
                ]
            ]
            array_for_prediction = np.array(list_for_array, dtype=np.float32)
            length = model.predict(array_for_prediction)[0]
            self.length.expression = application.modeler._arg_with_dim(length)
        else:  # pragma: no cover
            self.application.logger.warning("Machine learning algorithm aren't covered in IronPython.")



# -*- coding: utf-8 -*-

from plumbum.commands.base import (
    ERROUT,
    BaseCommand,
    ConcreteCommand,
    shquote,
    shquote_list,
)
from plumbum.commands.modifiers import (
    BG,
    FG,
    NOHUP,
    RETCODE,
    TEE,
    TF,
    ExecutionModifier,
    Future,
)
from plumbum.commands.processes import (
    CommandNotFound,
    ProcessExecutionError,
    ProcessLineTimedOut,
    ProcessTimedOut,
    run_proc,
)
